import unittest

import numpy as np
from featurebox.symbol.functions.dimfunc import Dim, dless, dnan, dim_map


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.a = Dim([1, 2, 3, 4, 5, 6, 7])
        self.b = Dim([2, 2, 3, 4, 5, 6, 7])

        self.dl = dless
        self.dn = dnan

        self.c = Dim([[1, 2, 3, 4, 5, 6, 7], [1, 2, 3, 4, 5, 6, 7]])
        self.dl2 = Dim([[0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0]])
        self.dn2 = Dim([[np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                        [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]])

        self.dim_map = dim_map()

    def test_add(self):
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn

        assert a != b
        assert a != dl
        assert a != dn
        assert a == c
        self.assertEqual(a, a + 1)
        self.assertEqual(a, 1 + a)
        self.assertEqual(a, a + a)

        # assert (a + b).anyisnan()
        self.assertNotEqual(a, a + dl)
        self.assertNotEqual(a, dl + a)
        assert (a + dn).anyisnan()
        assert (dn + a).anyisnan()

        self.assertEqual(2, (1 + c).ndim)
        self.assertEqual(2, (c + 1).ndim)
        self.assertEqual(1, (1 + a).ndim)
        self.assertEqual(1, (a + 1).ndim)
        self.assertEqual(2, (a + c).ndim)
        self.assertEqual(1, (a + b).ndim)
        self.assertEqual(2, (c + dl).ndim)
        self.assertEqual(2, (c + b).ndim)
        self.assertEqual(2, (c + dn).ndim)
        self.assertEqual(2, (dn + c).ndim)
        self.assertEqual(1, (dn + dn).ndim)

    def test_mul(self):
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn
        self.assertEqual(a, a * 1)
        self.assertEqual(a, 1 * a)
        self.assertEqual(Dim(np.array(a) * 2), a * a)

        self.assertEqual(Dim(np.array([3, 4, 6, 8, 10, 12, 14])), a * b)
        self.assertNotEqual(Dim(np.array([2, 4, 6, 8, 10, 12, 14])), a * b)
        self.assertEqual(a, a * dl)
        self.assertEqual(a, dl * a)
        assert (a * dn).anyisnan()
        assert (dn * a).anyisnan()

        self.assertEqual(2, (1 * c).ndim)
        self.assertEqual(2, (c * 1).ndim)
        self.assertEqual(1, (1 * a).ndim)
        self.assertEqual(1, (a * 1).ndim)
        self.assertEqual(2, (a * c).ndim)
        self.assertEqual(1, (a * b).ndim)
        self.assertEqual(2, (c * dl).ndim)
        self.assertEqual(2, (c * b).ndim)
        self.assertEqual(2, (c * dn).ndim)
        self.assertEqual(2, (dn * c).ndim)
        self.assertEqual(1, (dn * dn).ndim)
        self.assertEqual(1, (dn * dl).ndim)
        self.assertEqual(1, (b * dl).ndim)

    #
    def test_div(self):
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn
        self.assertEqual(a, a / 1)
        self.assertEqual(Dim(-np.array(a)), 1 / a)
        self.assertEqual(dless, a / a)
        self.a = Dim([1, 2, 3, 4, 5, 6, 7])
        self.b = Dim([2, 2, 3, 4, 5, 6, 7])
        self.assertEqual(Dim(np.array([-1, 0, 0, 0, 0, 0, 0])), a / b)
        self.assertNotEqual(Dim(np.array([2, 4, 6, 8, 10, 12, 14])), a / b)
        self.assertEqual(a, a / dl)
        self.assertEqual(Dim(-np.array(a)), dl / a)
        assert (a / dn).anyisnan()
        assert (dn / a).anyisnan()
        assert (dn / dn).anyisnan()

        self.assertEqual(2, (1 / c).ndim)
        self.assertEqual(2, (c / 1).ndim)
        self.assertEqual(1, (1 / a).ndim)
        self.assertEqual(1, (a / 1).ndim)
        self.assertEqual(2, (a / c).ndim)
        self.assertEqual(1, (a / b).ndim)
        self.assertEqual(2, (c / dl).ndim)
        self.assertEqual(2, (c / b).ndim)
        self.assertEqual(2, (c / dn).ndim)
        self.assertEqual(2, (dn / c).ndim)
        self.assertEqual(1, (dn / dn).ndim)
        self.assertEqual(1, (dn / dl).ndim)
        self.assertEqual(1, (b / dl).ndim)

    #
    def test_pow(self):
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn
        self.assertEqual(Dim(np.array(a)* 2), a ** 2)
        self.assertEqual(Dim(np.array(a) / 2), a ** 0.5)
        assert (4 ** a).anyisnan()
        assert (a ** b).anyisnan()

        self.assertEqual(1, (1 ** c).ndim)
        self.assertEqual(2, (c ** 1).ndim)
        self.assertEqual(1, (1 ** a).ndim)
        self.assertEqual(1, (a ** 1).ndim)
        self.assertEqual(1, (a ** c).ndim)
        self.assertEqual(1, (a ** b).ndim)
        self.assertEqual(1, (c ** dl).ndim)
        self.assertEqual(1, (c ** b).ndim)
        self.assertEqual(1, (c ** dn).ndim)
        self.assertEqual(1, (dn ** c).ndim)
        self.assertEqual(1, (dn ** dn).ndim)
        self.assertEqual(1, (dn ** dl).ndim)
        self.assertEqual(1, (b ** dl).ndim)

    def test_abs(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn
        my_funcs = self.dim_map
        func = my_funcs["Abs"]
        self.assertEqual(a, func(a))
        self.assertEqual(b, func(b))
        self.assertNotEqual(dl, func(4))

        self.assertEqual(2, func(c).ndim)
        self.assertEqual(2, func(a + c).ndim)
        self.assertEqual(2, func(c + b).ndim)
        self.assertEqual(1, func(b).ndim)

    #
    def test_exp(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn = self.a, self.b, self.c, self.dl, self.dn
        my_funcs = self.dim_map
        func = my_funcs["exp"]
        assert func(a).anyisnan()
        assert func(b).anyisnan()
        assert func(dn).anyisnan()
        self.assertNotEqual(dl, func(4))
        self.assertEqual(dl, func(dl))

        self.assertEqual(2, func(c).ndim)
        self.assertEqual(2, func(a + c).ndim)
        self.assertEqual(2, func(c + b).ndim)
        self.assertEqual(1, func(dl).ndim)
        self.assertEqual(1, func(a).ndim)

    def test_Flat(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn ,dl2,dn2= self.a, self.b, self.c, self.dl, self.dn,self.dl2,self.dn2
        my_funcs = self.dim_map
        func = my_funcs["MAdd"]

        self.assertEqual(a, func(a))
        self.assertEqual(b, func(b))
        self.assertEqual(dl, func(dl))
        assert func(dn).anyisnan()
        assert func(dn2).anyisnan()
        self.assertEqual(c, func(c))
        self.assertEqual(dl2, func(dl2))
        self.assertNotEqual(dl, func(3))

        self.assertEqual(1, func(a).ndim)
        self.assertEqual(1, func(b).ndim)
        self.assertEqual(1, func(dl).ndim)
        self.assertEqual(1, func(dn).ndim)
        self.assertEqual(1, func(c).ndim)

    def test_Diff(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn ,dl2,dn2= self.a, self.b, self.c, self.dl, self.dn,self.dl2,self.dn2
        c5 = a.copy()
        c5 = Dim(np.array([c5] * 5))

        my_funcs = self.dim_map
        func = my_funcs["MSub"]

        self.assertEqual(a, func(a))
        self.assertEqual(b, func(b))
        self.assertEqual(dl, func(dl))
        assert func(dn).anyisnan()
        assert func(dn2).anyisnan()
        self.assertEqual(c, func(c))
        self.assertEqual(dl2, func(dl2))
        self.assertNotEqual(dl, func(3))

        self.assertEqual(1, func(a).ndim)
        self.assertEqual(1, func(b).ndim)
        self.assertEqual(1, func(dl).ndim)
        self.assertEqual(1, func(dn).ndim)
        self.assertEqual(1, func(c).ndim)
        self.assertEqual(2, func(c5).ndim)

    def test_comp(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn ,dl2,dn2= self.a, self.b, self.c, self.dl, self.dn,self.dl2,self.dn2
        c5 = a.copy()
        c5 = Dim(np.array([c5] * 5))

        my_funcs = self.dim_map
        func = my_funcs["MMul"]

        self.assertEqual(a, func(a))
        self.assertEqual(b, func(b))
        self.assertEqual(dl, func(dl))
        assert func(dn).anyisnan()
        assert func(dn2).anyisnan()
        self.assertEqual(dl2**2, func(dl2))
        self.assertEqual(Dim(2 * np.array(c)), func(c))
        self.assertEqual(Dim(5 * np.array(c)), func(c5))
        self.assertNotEqual(dl, func(3))

        self.assertEqual(1, func(a).ndim)
        self.assertEqual(1, func(b).ndim)
        self.assertEqual(1, func(dl).ndim)
        self.assertEqual(1, func(dn).ndim)
        self.assertEqual(1, func(c).ndim)
        self.assertEqual(1, func(c5).ndim)

    #
    def test_quot(self):
        # my_funcs = {"Abs": my_abs, "exp": my_exp, "log": my_log, 'cos': my_cos, 'sin': my_sin,
        #             'sqrt': my_sqrt, "Flat": my_flat, "Comp": my_comp, "Diff": my_diff,
        #             "Quot": my_quot, "Self": my_self}
        a, b, c, dl, dn ,dl2,dn2= self.a, self.b, self.c, self.dl, self.dn,self.dl2,self.dn2
        c5 = a.copy()
        c5 = Dim(np.array([c5] * 5))

        my_funcs = self.dim_map
        func = my_funcs["MDiv"]

        self.assertEqual(a, func(a))
        self.assertEqual(b, func(b))
        self.assertEqual(dl, func(dl))
        assert func(dn).anyisnan()
        assert func(dn2).anyisnan()
        self.assertEqual(dl2, func(dl2))
        self.assertEqual(dless, func(c))
        self.assertEqual(c5, func(c5))
        self.assertNotEqual(dl, func(3))

        self.assertEqual(1, func(a).ndim)
        self.assertEqual(1, func(b).ndim)
        self.assertEqual(1, func(dl).ndim)
        self.assertEqual(1, func(dn).ndim)
        self.assertEqual(1, func(c).ndim)
        self.assertEqual(2, func(c5).ndim)

    def test_all(self):
        a, b, c, dl, dn,dl2= self.a, self.b, self.c, self.dl, self.dn,self.dl2
        c5 = a.copy()
        c5 = Dim(np.array([c5] * 5))

        my_funcs = self.dim_map
        MMul = my_funcs["MMul"]
        MDiv = my_funcs["MDiv"]
        ln= my_funcs["ln"]
        assert  ln(MDiv(dl2+ MMul(c))).anyisnan()
        assert MDiv(dl2+ MMul(c)).anyisnan()
        assert (dl2+ MMul(c)).anyisnan()


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn.functional as F
from torch import nn


class GaussianBlur(nn.Module):
    def __init__(self):
        super(GaussianBlur, self).__init__()
        kernel = [[0.03797616, 0.044863533, 0.03797616],
                  [0.044863533, 0.053, 0.044863533],
                  [0.03797616, 0.044863533, 0.03797616]]
        kernel = torch.FloatTensor(kernel).unsqueeze(0).unsqueeze(0)
        self.weight = nn.Parameter(data=kernel, requires_grad=False)

    def forward(self, x):
        x1 = x[:, 0]
        x2 = x[:, 1]
        x3 = x[:, 2]
        x1 = F.conv2d(x1.unsqueeze(1), self.weight, padding=2)
        x2 = F.conv2d(x2.unsqueeze(1), self.weight, padding=2)
        x3 = F.conv2d(x3.unsqueeze(1), self.weight, padding=2)
        x = torch.cat([x1, x2, x3], dim=1)
        return x



import numpy as np
import torch
from matplotlib.pyplot import imshow
from torch import nn

from featurebox.cam.cam import GradCRM
from featurebox.cam.fig import trans_to_numpy, visualize, trans_to_tensor, trans_to_tensors
from featurebox.data.datasets import CAMData


class VGG(nn.Module):

    def __init__(self, features, num_classes=1000, init_weights=True):
        super(VGG, self).__init__()
        self.features = features
        self.avgpool = nn.AdaptiveAvgPool2d((3, 3))
        self.classifier = nn.Sequential(
            nn.Linear(57600
                      , 1000),
            nn.ReLU(True),
            nn.Dropout(),
            # nn.Linear(1000, 1000),
            # nn.ReLU(True),
            # nn.Dropout(),
            nn.Linear(1000, 1),
        )
        if init_weights:
            self._initialize_weights()

    def forward(self, x):
        x = self.features(x)
        # x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)


def make_layers(cfg, batch_norm=False):
    layers = []
    in_channels = 1
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
            else:
                layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    return nn.Sequential(*layers)


if __name__ == "__main__":

    sample = 100
    torch.set_default_tensor_type(torch.FloatTensor)
    camdata = CAMData(100, (30, 30), random_state=1, question="reg", inter=False, re_range=(-0.5, 1))
    data = camdata.gen_data(featurize_x="total", self_func="de_2Dsite")
    data.x = camdata.cp_channel_to1(data.x)[0]
    x = data.x

    # y = data.y
    feature = data.feature
    y = torch.tensor([f[1] for f in feature]).reshape((-1, 1))

    t_x = trans_to_tensors(x)
    t_y = trans_to_tensor(y, channel=False)


    def showsite(t_x, feature, i):
        x = t_x[i]
        sitei = feature[i][0]
        x = trans_to_numpy(x)
        if x.ndim == 2:
            x = np.expand_dims(x, axis=-1)
            x = np.repeat(x, 3, axis=-1) / 2
        elif x.ndim == 3:
            if x.shape[-1] == 1:
                x = np.repeat(x, 3, axis=-1) / 2
        assert x.shape[-1] == 3
        x[..., 0][sitei[0]] = 1
        x[..., 0][sitei[1]] = 1
        x[..., 0][sitei[2]] = 1

        return x


    #

    # Max池化层,conv2d3kernel  输出通道,
    cfgs = {
        'A': [64, 64],
        'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
        'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
        'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
    }


    def vgg_self(cfg="A", batch_norm=False):

        model = VGG(make_layers(cfgs[cfg], batch_norm=batch_norm), num_classes=100)

        return model


    model = vgg_self(cfg="A", batch_norm=False)
    target_layer = model.features[-1]
    opitmizer = torch.optim.SGD(model.parameters(), lr=0.03)
    loss_fun = nn.MSELoss()
    for i in range(5):
        predict = model(t_x)
        # print(predict)
        loss = loss_fun(predict, t_y)
        loss = torch.as_tensor(loss, dtype=torch.float)

        print(float(loss))

        opitmizer.zero_grad()
        loss.backward()
        opitmizer.step()
    # # In [7]:
    # # # the target layer you want to visualize
    # target_layer = model.features[-1].conv2
    # # #
    # # # wrapper for class activation mapping. Choose one of the following.
    # wrapped_model = CAM(model, target_layer)
    wrapped_model = GradCRM(model, target_layer)
    # wrapped_model = GradCAMpp(model, target_layer)
    # wrapped_model = SmoothGradCAMpp(model, target_layer, n_samples=5, stdev_spread=0.15)
    # In [8]:
    img = t_x[7].unsqueeze(0)
    cam, idx = wrapped_model(img, idx=7)
    heatmap = visualize(img, cam)
    hm = (heatmap.squeeze().numpy().transpose(1, 2, 0))
    xp = showsite(t_x, feature, 7)
    imshow(xp)

    cam = cam.squeeze(0).numpy().transpose(1, 2, 0)
    cam = np.repeat(cam, 3, axis=2)
    imshow(cam)



#!/usr/bin/python
# coding:utf-8

# @author: wangchangxin
# @contact: 986798607@qq.com
# @software: PyCharm
# @file: preference.py
# @time: 2020/5/26 19:50

from itertools import combinations

import numpy as np
from numpy import random
from numpy.core import numeric


class PreMap(numeric.ndarray):

    def __new__(cls, data):

        assert isinstance(data, numeric.ndarray)
        dtype = np.float32
        arr = numeric.array(data, dtype=dtype, copy=True)
        shape = arr.shape
        ret = numeric.ndarray.__new__(cls, shape, dtype=np.float32,
                                      buffer=arr,
                                      order='c')
        return ret

    @classmethod
    def from_shape(cls, shape):
        """

        Parameters
        ----------
        shape:int
            shape of premap

        Returns
        -------
        PreMap
        """
        shape = (shape, shape)
        ret = numeric.ndarray.__new__(cls, shape, dtype=np.float16)
        val = np.full(shape, (shape[0] - 0.01) / (shape[0] * (shape[0] - 1)), dtype=np.float16)
        ret[:] = val
        for i in range(ret.shape[0]):
            ret[i, i] = 0.01 / shape[0]
        return cls(ret)

    def down_other_point(self, *sv):

        """
        Use for binding.rate the others and add the subbed value to the [index1,index2]
        the rate are [0,1)
            
        Parameters
        ----------
        sv:[index1,index2,rate]
            site to set value

        """
        e = self.copy()
        a = sv[0]
        b = sv[1]
        c = sv[2]
        if c == 1:
            c -= 0.0001

        self[(a, b), :] *= (1 - c)
        self[:, (a, b)] *= (1 - c)
        self[a, a] /= (1 - c)
        self[b, b] /= (1 - c)
        self[a, b] /= (1 - c) ** 2
        self[b, a] /= (1 - c) ** 2
        nv = e - self
        self[a, b] = (np.sum(nv) + nv[a, a] + nv[b, b]) / 4 + e[a, b]
        self[b, a] = self[a, b]

        decend = np.sum(nv[(a, b), :], axis=0)
        ratio = decend / np.sum(e, axis=0)

        e_add = e * (1 + ratio)
        e_add[(a, b), :] = self[(a, b), :]
        e_add[:, (a, b)] = self[:, (a, b)]

        self[:] = e_add

    def set_sigle_point(self, *sv):
        """
        set the value of [index1,index2]
        the rate are [0,1)
            
        Parameters
        ----------
        sv:[index1,index2,rate]
            site to set value


        """
        a = sv[0]
        b = sv[1]
        c = sv[2]
        if c == 1:
            c -= 0.001
        self[a, b] = c
        self[b, a] = c

    def set_ratio(self, *sv):
        """
        Rate the [index1,index2] to sum and add the subbed value to the others.
        under check.
        Parameters
        
        ----------
        sv:[index1,index2,rate]
            rate in [0,n), if [0,1) down, if [1,n) up
        """

        a = sv[0]
        b = sv[1]
        c = sv[2]
        if c == 1:
            c -= 0.001

        suma = np.sum(self[a])
        sumb = np.sum(self[b])
        st = c * (suma + sumb) / 2
        vale = st - self[a, b]
        base = (suma + sumb - 2 * self[a, b]) / 2
        coef = 1 - vale / base

        self[(a, b), :] *= coef
        self[:, (a, b)] *= coef
        self[a, a] /= coef
        self[b, b] /= coef

        self[a, b] = st
        self[b, a] = st

    def set_ratio_point(self, *sv):
        """
        Rate the [index1,index2] to self and add the subbed value to the others.
        under check.
        Parameters
        
        ----------
        sv:[index1,index2,rate]
            [0,n)
            [0,1) down,
            [1,n) up
        """

        a = sv[0]
        b = sv[1]
        c = sv[2]
        if c == 1:
            c -= 0.00001

        st = c * self[a, b]
        vale = (1 - c) * self[a, b]
        base = (np.sum(self[a]) + np.sum(self[b]) - 2 * self[a, b]) / 2
        coef = 1 + vale / base

        self[(a, b), :] *= coef
        self[:, (a, b)] *= coef
        self[a, a] /= coef
        self[b, b] /= coef

        self[a, b] = st
        self[b, a] = st

    def noise(self):
        """add noise with 1% scale"""
        ran = random.random_sample(size=self.shape)
        rant = ran.T
        noise = ran + rant - 1
        self[:] = self * (1 + noise * 0.1)

    def add_new(self):
        """add new features to self"""
        npself = self
        ave = npself.mean(axis=0)
        se = np.concatenate((npself, ave.reshape(-1, 1)), axis=1)
        ave = np.append(ave, 0.0)
        se = np.concatenate((se, ave.reshape(1, -1)), axis=0)
        se *= (1 - 1 / se.shape[0])

        return PreMap(se)

    def update(self, ind, pset, ratio=0.5):
        """

        Parameters
        ----------
        ind: SymbolTree
            individual
        pset: SymbolSet
            SymbolSet
        ratio: float [0,1]
            change ratio
            
        Returns
        -------
        self
        """
        ratio = ratio / self.shape[0]
        ters = pset.terminals_and_constants
        pri = [prii for prii in ind if prii.arity == 0]
        indexs = [ters.index(prii) for prii in pri]
        iters = combinations(indexs, 2)
        se = self.copy()
        [self.down_other_point(i, j, ratio) for i, j in iters]
        self[:] += se
        self[:] /= 2
        self.noise()

    def get_indexes_value(self, indexes, weight=None):
        """

        Parameters
        ----------
        indexes:tuple, indexes
            get the value of average of indexes affect
        weight: None, tuple,np.ndarray
            the same size with self.shape[0]
            
        Returns
        -------
        probability list
        """
        if isinstance(indexes, int):
            indexes = (indexes,)
        indexes = tuple(indexes)
        var = self[indexes, :]
        if var.ndim == 2:
            if weight is not None:
                weight = np.array(weight)
                assert var.shape[0] == weight.shape[0]
            var = np.average(var, weights=weight, axis=0)

        return var / (np.sum(var))

    def get_ind_value(self, ind, pset):
        """
        get the value according to ind

        Parameters
        ----------
        ind: SymbolTree
        pset:SymbolSet
        
        Returns
        ----------
        probability list
        """
        ters = pset.terminals_and_constants
        pri = [prii for prii in ind if prii.arity == 0]
        if len(pri) >= 1:
            indexes = [ters.index(i) for i in pri]

            return self.get_indexes_value(indexes, weight=None)
        else:
            return None

    def get_one_node_value(self, ind=None, pset=None, node=None, site=None, ):
        """
        get affect value except site node.
        
        Parameters
        ----------
        ind: SymbolTree
        pset:SymbolSet
        node:Terminals
        site:site of Terminals

        Returns
        -------
        probability list
        """
        if not node:
            node = ind[site]

        ters = pset.terminals_and_constants

        pri = [prii for prii in ind if prii.arity == 0]
        if len(pri) > 1:
            weight = list(range(len(pri)))

            if site:
                ast = len([k for k, j in enumerate(ind[:site]) if node == j])
            else:
                ast = 0
            loc = [k for k, j in enumerate(pri) if node == j][ast]  # choice first 0ne

            weight = np.array([len(weight) - abs(i - loc) for i in weight])
            weight = weight / weight.sum(keepdims=True)
            indexes = [ters.index(prii) for prii in pri]
            indexes = np.delete(indexes, loc)
            weight = np.delete(weight, loc)

            prob = self.get_indexes_value(indexes, weight=weight)
        else:
            prob = None
        return prob

    def get_nodes_value(self, ind=None, pset=None, node=None, site=None, ):
        """
        get affect value except sites nodes.
        
        Parameters
        ----------
        ind: SymbolTree
        pset:SymbolSet
        node:Terminals
        site:site of Terminals

        Returns
        -------
        probability list
        """
        if isinstance(site, int):
            site = (site,)
        if isinstance(node, int):
            node = (node,)

        if not node:
            node = [ind[i] for i in site]

        ters = pset.terminals_and_constants

        pri = [prii for prii in ind if prii.arity == 0]
        if len(pri) > 1:
            weight = list(range(len(pri)))
            if site is not None:
                ast = [len([k for k, j in enumerate(ind[:sitei]) if nodei == j]) for nodei, sitei in zip(node, site)]
            else:
                ast = [0] * len(node)
            locs = [[k for k, j in enumerate(pri) if nodei == j][asti] for nodei, asti in
                    zip(node, ast)]  # choice first 0ne

            weight = np.array([[len(weight) - abs(i - loci) for i in weight] for loci in locs])

            if weight.ndim == 1:
                weight = weight / weight.sum()
            elif weight.ndim == 2:
                weight = weight.mean(axis=0)
            else:
                pass
            indexes = [ters.index(prii) for prii in pri]
            indexes = np.delete(indexes, locs)
            weight = np.delete(weight, locs)

            prob = self.get_indexes_value(indexes, weight=weight)
        else:
            prob = None
        return prob

# if __name__ == "__main__":
#    # import copy
#    # from featurebox.symbol.gp import mutNodeReplacementVerbose, mutDifferentReplacementVerbose
#    # from numpy import random
#    # from sklearn.datasets import load_boston
#    #
#    # from featurebox.symbol.base import SymbolSet
#    # from featurebox.symbol.base import SymbolTree
#    # from featurebox.symbol.dim import dless, Dim
#    #
#    # random.seed(3)
#    # # data
#    # data = load_boston()
#    # x = data["data"]
#    # y = data["target"]
#    # c = [6, 3, 4]
#    # # unit
#    # from sympy.physics.units import kg
#    #
#    # x_u = [kg] * 13
#    # y_u = kg
#    # c_u = [dless, dless, dless]
#    #
#    # x, x_dim = Dim.convert_x(x, x_u, target_units=None, unit_system="SI")
#    # y, y_dim = Dim.convert_xi(y, y_u)
#    # c, c_dim = Dim.convert_x(c, c_u)
#    #
#    # # symbolset
#    # pset0 = SymbolSet()
#    # pset0.add_features(x, y, x_dim=x_dim, y_dim=y_dim, group=[[1, 2], [4, 5]])
#    # pset0.add_constants(c, dim=c_dim, prob=None)
#    # pset0.add_operations(power_categories=(2, 3, 0.5),
#    #                      categories=("Add", "Mul", "Neg", "Abs"),
#    #                      self_categories=None)
#    # pset0.personal_maps([[0,1,0.9]])
#    # pset0.personal_maps([[0,5,0.7]])
#    # pp = np.array(pset0.premap)
#    # for i in range(100):
#    #     a = SymbolTree.genGrow(pset0, 2, 4, per=True)
#    #     mutNodeReplacementVerbose(a, pset=pset0)
#    #     mutDifferentReplacementVerbose(a, pset=pset0)
#    # print(np.sum(pset0.premap))
#    # pset0.personal_preference([[3, 4, 0.8]])
#    # print(np.sum(pset0.premap))
#    # pset0.add_tree_to_features(a)
#    # print(np.sum(pset0.premap))
#    # pset0.premap.noise()
#    # print(np.sum(pset0.premap))
#    # values = pset0.premap.get_indexes_value(4)
#    # values2 = pset0.premap.get_ind_value(a, pset0)
#    premap = PreMap.from_shape(10)
#    premap.down_other_point(*[0, 1, 1])
#    premap.down_other_point(*[0, 2, 0.50])
#    premap.down_other_point(*[0, 3, 0.33])
#    premap.down_other_point(*[0, 4, 0.25])
#    # premap.down_other_point(*[0, 3, 0.5])
#    # premap.set_ratio(*[0, 1, 1])
#    # premap.set_ratio(*[0, 3, 0.5])
#    # premap.set_ratio(*[0, 5, 0.5])
#    # premap.down_others(*[0, 2, 0.25])
#    pp = np.array(premap)
#
#    sums1 = np.sum(premap, axis=0)
#    sums2 = np.sum(premap, axis=1)



# -*- coding: utf-8 -*-

# @Time    : 2020/1/6 9:41
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
from pymatgen.io.vasp import inputs

from featurebox.tools.imports import BatchFile

# Kpoints
# Poscar
# Incar
#
# VaspInput
# Chgcar
#
# Elfcar
# Oszicar
# Outcar

if __name__ == "__main__":
    a = BatchFile(r"C:\Users\Administrator\Desktop\file")
    # a.filter_dir_name("line",layer=-1)
    a.filter_dir_name(exclud="band_line")
    a.filter_file_name("POSCAR")
    lists = a.merge()
    # a.to_path(r"C:\Users\Administrator\Desktop\newss", add_dir=[-3, -2, -1], flatten=False)

    # ii=inputs.Incar
    # ii=inputs.Kpoints
    ii = inputs.Poscar

    for i, j in enumerate(lists):
        isi = ii.from_file(lists[0])
        # print(ii)



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/29 19:47
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause


"""
# Just a copy from xenonpy
for pictures use skimage.io.ImageCollection
"""
import copy
import glob
import os
import re
import shutil
from collections import defaultdict
from functools import partial
from pathlib import Path
from warnings import warn

import joblib
import pandas as pd
import requests
from skimage import io

from featurebox.tools.exports import def_pwd


class Call(object):
    """
    call file in paths
    """

    @staticmethod
    def extension(index_col=0):
        read_csv = partial(pd.read_csv, index_col=index_col)
        read_excel = partial(pd.read_excel, index_col=index_col)
        extension = dict(
            pkl_pd=('pkl_pd', pd.read_pickle),
            csv=('csv', read_csv),
            xlsx=('xlsx', read_excel),
            pkl_sk=('pkl_sk', joblib.load),
            png=("png", io.imread),
            jpg=("jpg", io.imread),
        )
        return extension

    __re__ = re.compile(r'[\s\-.]')

    def __init__(self, *paths, backend='pkl_pd', prefix_with_upper=None, index_col=0):
        """
        :param backend:default imported type to show
        :param prefix_with_upper: prefix_with_upper for all file add to file in this code to excape same name
        :type paths: None, ..|data_cluster, or F:data_cluster|data1
        """
        self._backend = backend
        self.index_col = index_col
        self._files = None
        self.__extension__ = self.extension(index_col)

        if len(paths) == 0:
            self._paths = ('.',)
        else:
            self._paths = paths

        if not prefix_with_upper:
            prefix_with_upper = ()
        self._prefix = prefix_with_upper

        self._make_index(prefix_with_upper=prefix_with_upper)

    def _make_index(self, *, prefix_with_upper):
        def make(path_):
            patten = self.__extension__[self._backend][0]
            files = glob.glob(str(path_ / ('*.' + patten)))

            def _nest(_f):
                f_ = _f
                return lambda s: s.__extension__[s._backend][1](f_)

            for f in files:
                # selection data_cluster
                f = Path(f).resolve()
                parent = re.split(r'[\\/]', str(f.parent))[-1]
                # parent = str(f.parent).split('\\/')[-1]
                fn = f.name[:-(1 + len(patten))]
                fn = self.__re__.sub('_', fn)
                if prefix_with_upper:
                    fn = '_'.join([parent, fn])

                if fn in self._files:
                    warn("file %s with x_name %s already bind to %s and will be ignored" %
                         (str(f), fn, self._files[fn]), RuntimeWarning)
                else:
                    self._files[fn] = str(f)
                    setattr(self.__class__, fn, property(_nest(str(f))))

        self._files = defaultdict(str)
        for path in self._paths:
            path = Path(path).expanduser().absolute()
            if not path.exists():
                raise RuntimeError('%s not exists' % str(path))
            make(path)

    @classmethod
    def from_http(cls, url, save_to, *, filename=None, chunk_size=256 * 1024, params=None,
                  **kwargs):
        """
        Get file object via a http request.

        Parameters
        ----------
        url: str
            The resource url.
        save_to: str
            The path of a path to save the downloaded object into it.
        filename: str, optional
            Specific the file x_name when saving.
            Set to ``None`` (default) to use a inferred x_name from http header.
        chunk_size: int, optional
            Chunk size.
        params: any, optional
            Parameters will be passed to ``requests.get`` function.
            See Also: `requests <http://docs.python-requests.org/>`_
        kwargs: dict, optional
            Pass to ``requests.get`` function as the ``kwargs`` parameters.

        Returns
        -------
        str
            File path contains file x_name.
        """
        r = requests.get(url, params, **kwargs)
        r.raise_for_status()

        if not filename:
            if 'filename' in r.headers:
                filename = r.headers['filename']
            else:
                filename = url.split('/')[-1]

        if isinstance(save_to, str):
            save_to = Path(save_to)
        if not isinstance(save_to, Path) or not save_to.is_dir():
            raise RuntimeError('%s is not a legal path or not point to a path' % save_to)

        file_ = str(save_to / filename)
        with open(file_, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return file_

    def __repr__(self):
        cont_ls = ['<{}> includes:'.format(self.__class__.__name__)]

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, v))

        return '\n'.join(cont_ls)

    def csv(self):
        return Call(*self._paths, backend='csv', prefix_with_upper=self._prefix, index_col=self.index_col)

    def pickle_pd(self):
        return Call(*self._paths, backend='pkl_pd', prefix_with_upper=self._prefix, index_col=self.index_col)

    def pickle_sk(self):
        return Call(*self._paths, backend='pkl_sk', prefix_with_upper=self._prefix, index_col=self.index_col)

    def xlsx(self):
        return Call(*self._paths, backend='xlsx', prefix_with_upper=self._prefix, index_col=self.index_col)

    def png(self):
        return Call(*self._paths, backend='png', prefix_with_upper=self._prefix, index_col=self.index_col)

    def jpg(self):
        return Call(*self._paths, backend='jpg', prefix_with_upper=self._prefix, index_col=self.index_col)

    def __call__(self, *args, **kwargs):
        return self.__extension__[self._backend][1](*args, **kwargs)

    def __getattr__(self, name):
        """
        Returns sub-dataset.

        Parameters
        ----------
        name: str
            Dataset x_name.

        Returns
        -------
        spath
        """
        if name in self.__extension__:
            return self.__class__(*self._paths, backend=name, prefix=self._prefix)
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))


def check_file(spath, file_path, suffix=None):
    os.chdir(file_path)
    # print(os.path.abspath(os.curdir))
    all_file = os.listdir()
    files = []
    for f in all_file:
        if os.path.isdir(f):
            files.extend(check_file(spath, file_path + '\\' + f, suffix=suffix))
            os.chdir(file_path)
        else:
            if not suffix:
                che = True
            elif suffix == "":
                che = "" == os.path.splitext(f)[1]
            else:
                che = "".join((".", suffix)) == os.path.splitext(f)[1]
            if che:
                rel_path = file_path.replace(spath, "")
                parents = re.split(r'[\\/]', str(rel_path))
                files.append([parents, f])
            else:
                pass
    return files


class BatchFile():
    """
    search files and filter files and re-site files
    """

    def __init__(self, path=None, suffix=None):

        path = def_pwd(path)
        self.path = path

        parents = re.split(r'[\\/]', str(path))
        self.parents = parents

        self.file_list = check_file(path, path, suffix=suffix)
        self.init_file = tuple(self.file_list)

    def filter_file_name(self, includ=None, exclud=None):
        file_list_filter = []
        for file_i in self.file_list:
            name = file_i[1]

            if includ and not exclud:
                if includ in name:
                    file_list_filter.append(file_i)
            elif not includ and exclud:
                if exclud not in name:
                    file_list_filter.append(file_i)
            elif includ and exclud:
                if includ in name and exclud not in name:
                    file_list_filter.append(file_i)
            else:
                raise TypeError("one of includ,exclud must be str")

        self.file_list = file_list_filter

    def filter_dir_name(self, includ=None, exclud=None, layer=-1):
        file_list_filter = []

        for file_i in self.file_list:
            if isinstance(layer, int):
                layer = [layer, ]

            if isinstance(layer, list):
                name = [file_i[0][i] for i in layer]
            else:
                name = file_i[0]
            name = "".join(name)

            if includ and not exclud:
                if includ in name:
                    file_list_filter.append(file_i)
            elif not includ and exclud:
                if exclud not in name:
                    file_list_filter.append(file_i)
            elif includ and exclud:
                if includ in name and exclud not in name:
                    file_list_filter.append(file_i)
            else:
                raise TypeError("one of includ,exclud must be str")

        self.file_list = file_list_filter

    def merge(self, path=None, flatten=False, add_dir="3-layer"):
        if not path:
            path = self.path
            flatten = False
        if not add_dir:
            add_dir = []
        elif add_dir == "3-layer":
            add_dir = [-1, -2, -3]
        if isinstance(add_dir, int):
            add_dir = [add_dir, ]

        file_list_merge = []
        for file_i in self.file_list:
            site = copy.copy(file_i[0])
            if isinstance(flatten, dict):
                site = [site[_] for _ in add_dir]
                site_c = ""
                for i, j in enumerate(site):
                    i -= len(site)
                    if i in flatten.keys():
                        if flatten[i] in [True, "layer", "dir", "folder", 1, "s"]:
                            site_c += "".join((j, "/"))
                        else:
                            site_c += "".join((j, "_"))
                    else:
                        site_c += "".join((j, "_"))
                site_c = re.split(r'[\\/]', str(site_c))
                site_c[-1] += file_i[1]
                file_list_merge.append(os.path.join(path, *site_c))

            elif flatten:
                site = [site[_] for _ in add_dir]
                site.append(file_i[1])
                site = "_".join(site)
                file_list_merge.append(os.path.join(path, site))
            else:
                site.append(file_i[1])
                file_list_merge.append(os.path.join(path, *site))

        return file_list_merge

    def to_path(self, new_path, flatten=True, add_dir="3-layer"):
        self.file_list_merge = self.merge()
        new_path = def_pwd(new_path)
        self.file_list_merge_new = self.merge(path=new_path, flatten=flatten, add_dir=add_dir)
        if len(set(self.file_list_merge_new)) < len(set(self.file_list_merge)):
            raise UserWarning("there are same name files after flatten folders. "
                              "please change add_dir to add difference prefix to files", )
        # parallelize(-1, shutil.copy, zip(self.file_list_merge, self.file_list_merge_new), respective=True)
        for i, j in zip(self.file_list_merge, self.file_list_merge_new):
            path_i = os.path.split(j)[0]
            if not os.path.exists(path_i):
                os.makedirs(path_i)
            shutil.copy(i, j)


# others please use shutil
# shutil.copytree()

if __name__ == "__main__":
    a = BatchFile(r"C:\Users\Administrator\Desktop\file")
    # a.filter_dir_name("line",layer=-1)
    a.filter_file_name("INCAR")
    a.to_path(r"C:\Users\Administrator\Desktop\newss", add_dir=[-3, -2, -1], flatten=False)



import operator
import unittest

from featurebox.symbol.functions.dimfunc import dless

from featurebox.symbol.base import CalculatePrecisionSet
from featurebox.symbol.base import SymbolSet
from featurebox.symbol.base import SymbolTree
from featurebox.symbol.gp import cxOnePoint, varAnd, genGrow, staticLimit, mutShrink, selKbestDim
from featurebox.tools.packbox import Toolbox


class MyTestgp(unittest.TestCase):

    def setUp(self):
        self.SymbolTree = SymbolTree
        self.pset = SymbolSet()

        from sklearn.datasets import load_boston

        data = load_boston()
        x = data["data"]
        y = data["target"]

        self.x = x
        self.y = y
        # self.pset.add_features(x, y, )
        self.pset.add_features(x, y, x_group=[[1, 2], [4, 5]])
        self.pset.add_constants([6, 3, 4], c_dim=[dless, dless, dless], c_prob=None)
        self.pset.add_operations(power_categories=(2, 3, 0.5),
                                 categories=("Add", "Mul", "Neg", "Abs"),
                                 self_categories=None)

        from sklearn.metrics import r2_score, mean_squared_error

        self.cp = CalculatePrecisionSet(self.pset, scoring=[r2_score, mean_squared_error],
                                        score_pen=[1, -1],dim_type=None,
                                        filter_warning=True)

    def test_gp_flow(self):
        from numpy import random
        random.seed(1)
        cpset = self.cp
        # def Tree
        from deap.base import Fitness
        from featurebox.tools import newclass
        Fitness_ = newclass.create("Fitness_", Fitness, weights=(1, -1))
        PTree_ = newclass.create("PTrees_", SymbolTree, fitness=Fitness_)

        # def selection
        toolbox = Toolbox()

        # toolbox.register("select", selTournament, tournsize=3)
        toolbox.register("select", selKbestDim, dim_type=dless)
        # selBest
        toolbox.register("mate", cxOnePoint)
        # def mutate
        toolbox.register("generate", genGrow, pset=cpset, min_=2, max_=3)
        # toolbox.register("mutate", mutUniform, expr=toolbox.generate, pset=cpset)
        # toolbox.register("mutate", mutNodeReplacement, pset=cpset)
        toolbox.register("mutate", mutShrink,pset=cpset)

        toolbox.decorate("mate", staticLimit(key=operator.attrgetter("height"), max_value=10))
        toolbox.decorate("mutate", staticLimit(key=operator.attrgetter("height"), max_value=10))
        # def elaluate

        # toolbox.register("evaluate", cpset.parallelize_calculate, n_jobs=4, add_coef=True,
        # inter_add=False, inner_add=False)

        # toolbox.register("parallel", parallelize, n_jobs=1, func=toolbox.evaluate, respective=False, tq=False)

        population = [PTree_.genGrow(cpset, 3, 4) for _ in range(10)]
        # si = sys.getsizeof(cpset)
        for i in range(5):
            invalid_ind = [ind for ind in population if not ind.fitness.valid]
            invalid_ind_score = cpset.parallelize_score(inds=invalid_ind)

            for ind, score in zip(invalid_ind, invalid_ind_score):
                ind.fitness.values = score[0]
                ind.y_dim = score[1]
            # si2 = sys.getsizeof(invalid_ind[0])
            # invalid_ind=[i.compress() for i in invalid_ind]
            # si3 = sys.getsizeof(invalid_ind[0])
            # print(si3,si2,si)
            population = toolbox.select(population, len(population))
            offspring = varAnd(population, toolbox, 1, 1)
            population[:] = offspring
            # cpsl.compress()


# if __name__ == '__main__':
#
#     unittest.main()

if __name__ == "__main__":
    import time

    a = time.time()
    se = MyTestgp()
    se.setUp()
    b = time.time()
    se.test_gp_flow()
    c = time.time()



# -*- coding: utf-8 -*-

# @Time   : 2019/5/19 16:25
# @Author : Administrator
# @Project : feature_preparation
# @FileName: neighbor.py
# @Software: PyCharm

"""
this is a description
"""

from abc import ABC

from featurebox.featurizers.base import BaseFeaturizer


class Neighborizer(BaseFeaturizer, ABC):

    # @property
    # def feature_labels(spath):
    #     return [str(d) for d in spath._interval[1:]]

    def __init__(self, tol=0.001, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        tol:
        n_jobs: int
            The number of jobs to run in parallel for both _fit and Fit. Set -1 to use all cpu cores (default).
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input0 type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

        self.tol = tol
        self.__authors__ = ['TsumiNa']

    def featurize(self, structure, r_max=7):
        """
        Args:
            structure: Pymatgen Structure object.
            r_max:
        """
        if not structure.is_ordered:
            raise ValueError("Disordered structure support not built yet")

        # Get the distances between all atoms
        neighbors_lst = structure.get_all_neighbors(r_max)

        neighbors_lst = [[(i[0], round(i[1], 4)) for i in neighbors_lsti] for neighbors_lsti in neighbors_lst]

        neighbors_lst = [[(i[0].specie, i[1]) for i in
                          sorted(neighbors_lsti, key=lambda s: s[1])] for neighbors_lsti in neighbors_lst]
        set01 = [sorted(set(neighbors_lsti), key=lambda s: s[1]) for neighbors_lsti in neighbors_lst]
        dict_count_all = []
        for seti, neighbors_lsti in zip(set01, neighbors_lst):
            dict_count = {}
            for item in seti:
                dict_count.update({item: (item[1], neighbors_lsti.count(item))})
            dict_count_all.append(dict_count)
        return dict_count_all



# -*- coding: utf-8 -*-

# @Time    : 2019/11/1 15:57
# @Email   : 986798607@qq.ele_ratio
# @Software: PyCharm
# @License: BSD 3-Clause

from abc import ABC

import numpy as np

from featurebox.featurizers.base import BaseFeaturizer
from featurebox.featurizers.extrastats import PropertyStats


class UnionFeaturizer(BaseFeaturizer, ABC):
    """
    transform method should input0 comp_index rather than entries
    """

    def __init__(self, comp, couple_data, couple=2, stats=("mean",)):
        super().__init__()
        self.couple = couple
        self.comp = comp
        self.elem_data = couple_data
        self.stats = stats
        # Initialize stats computer

    def featurize(self, comp_number=0):
        """
        Get elemental property attributes

        Args:
            comp: Pymatgen composition object

        Returns:
            all_attributes: Specified property statistics of features
            :param comp_number:
        """
        comp = self.comp[comp_number]
        elem_data = self.elem_data[comp_number].values

        # Get the element names and fractions
        elements, fractions = zip(*comp.element_composition.items())
        elem_data = np.reshape(elem_data, (self.couple, -1), order="F")
        all_attributes_all = []
        for stat in self.stats:
            all_attributes = [PropertyStats.calc_stat(elem_data_i, stat, fractions) for elem_data_i in elem_data.T]
            all_attributes_all.extend(all_attributes)
        return np.array(all_attributes_all)

    def feature_labels(self):
        """
        Generate attribute names.

        Returns:
            ([str]) attribute labels.
        """
        name = np.array(self.elem_data.columns.values)[::self.couple]
        name = [i.split("_")[0] + "%s" % j for i in name for j in self.stats]
        return name



#!/usr/bin/python
# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:13
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: GNU Lesser General Public License v3.0

import functools
import sympy
from sympy import Function



def sym_vector_map():
    """str to sympy.Expr function"""

    def Div(left, right):
        return left / right

    def Sub(left, right):
        return left - right

    def rem(ax):
        return 1 - ax

    functions1 = {"sin": sympy.sin, 'cos': sympy.cos, 'exp': sympy.exp, 'ln': sympy.ln,"log": sympy.ln,
                  'Abs': sympy.Abs, "Neg": functools.partial(sympy.Mul, -1.0),
                  "Rec": functools.partial(sympy.Pow, e=-1.0), "Self": lambda x: x,
                  "Rem": rem}
    functions2 = {"Add": sympy.Add, 'Sub': Sub, 'Mul': sympy.Mul, 'Div': Div}

    return functions1, functions2


def sym_dispose_map():
    """user's str to sympy.expr function"""
    Flat = Function("MAdd")
    Comp = Function("MMul")
    Diff = Function("MSub")
    Quot = Function("MDiv")
    Self = lambda x: x
    Conv = Function("Conv")
    Flat.is_jump = False
    Comp.is_jump = False
    Diff.is_jump = True
    Quot.is_jump = True
    Conv.is_jump = True
    Self.is_jump = True

    Flat.keep = False
    Comp.keep = False
    Diff.keep = False
    Quot.keep = False
    Conv.keep = True
    Self.keep = True

    return {"MAdd": Flat, "MMul": Comp, "MSub": Diff, "MDiv": Quot, "Conv": Conv, "Self": Self}




#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/27 16:57
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
# calculate the the correction of columns,or cal_group columns
"""
import copy
import random

import numpy as np
from sklearn.datasets import load_boston

from featurebox.selection.mutibase import MutiBase
from featurebox.tools.tool import name_to_name


class Corr(MutiBase):
    """
    _fit-->shrink_cof-->count_coef-->remove_coef
    """

    def __init__(self, threshold=0.85, muti_grade=2, muti_index=None, must_index=None):
        """

        :param threshold:
        :param muti_grade:cal_group size,calculate the correction between cal_group
        :param muti_index:the range of muti_grade
        :param must_index:the columns force to index
        """
        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.threshold = threshold
        self.cov = None
        self.cov_shrink = None
        self.shrink_list = []

    @staticmethod
    def MIC_matirx(dataframe, alpha=0.6, c=15):

        from minepy import MINE
        mine = MINE(alpha=alpha, c=c)
        data = np.array(dataframe)
        n = len(data[0, :])
        result = np.zeros([n, n])

        for i in range(n):
            for j in range(n):
                mine.compute_score(data[:, i], data[:, j])
                result[i, j] = mine.mic()
                result[j, i] = mine.mic()
        RT = result
        return RT

    def fit(self, data, pre_cal=None, method="mean", alpha=0.6, c=15):
        if pre_cal is None:

            cov = np.corrcoef(data, rowvar=False, )
        elif pre_cal is "mic":
            cov = self.MIC_matirx(data, alpha=alpha, c=c)

        elif isinstance(pre_cal, np.ndarray) and pre_cal.shape[0] == data.shape[1]:
            cov = pre_cal
        else:
            raise TypeError("pre_cal is None or coef of data_cluster with shape(data_cluster[0],data_cluster[0])")
        cov = np.nan_to_num(cov - 1) + 1
        self.cov = cov
        self.data = data
        self.shrink_list = list(range(self.cov.shape[0]))
        self._shrink_coef(method=method)

    def _shrink_coef(self, method="mean" or "max"):

        if self.check_muti:
            self.shrink_list = list(range(self.cov.shape[0]))
            self.shrink_list = list(self.feature_fold(self.shrink_list))

            cov = self.cov
            single = tuple([i for i in self.shrink_list if i not in self.check_muti])
            muti = tuple([i for i in self.shrink_list if i in self.check_muti])

            cov_muti_all = []
            le = self.muti_grade
            while le:
                index = []
                index.extend(single)
                index.extend([i + le - 1 for i in muti])
                index.sort()
                cov_muti_all.append(cov[index][:, index])
                le -= 1
            cov_muti_all = np.array(cov_muti_all)
            if method is "mean":
                cov_new = np.mean(cov_muti_all, axis=0)
            else:
                cov_new = np.max(cov_muti_all, axis=0)
            self.cov_shrink = cov_new
            return self.cov_shrink
        else:
            self.cov_shrink = self.cov

    def transform_index(self, data):
        if isinstance(data, int):
            return self.shrink_list.index(data)
        elif isinstance(data, (list, tuple)):
            return [self.shrink_list.index(i) for i in data]

    def inverse_transform_index(self, data):
        if isinstance(data, int):
            return self.shrink_list[data]
        elif isinstance(data, (list, tuple)):
            return [self.shrink_list[i] for i in data]
        else:
            pass

    def transform(self, data):
        if isinstance(data, (list, tuple)):
            return data[self.shrink_list]
        elif isinstance(data, np.ndarray) and data.ndim == 1:
            return data[self.shrink_list]
        elif isinstance(data, np.ndarray) and data.ndim == 2:
            return data[:, self.shrink_list]
        else:
            pass

    def count_cof(self, cof=None):

        """check cof and count the number"""
        if cof is None:
            cof = self.cov_shrink
        if cof is None:
            raise NotImplemented("imported cof is None")

        list_count = []
        list_coef = []
        g = np.where(abs(cof) >= self.threshold)
        for i in range(cof.shape[0]):
            e = np.where(g[0] == i)
            com = list(g[1][e])
            # ele_ratio.remove(i)
            list_count.append(com)
            list_coef.append([cof[i, j] for j in com])
        self.list_coef = list_coef
        self.list_count = list_count
        return list_coef, list_count

    @staticmethod
    def remove_coef(cof_list_all):
        """delete the index of feature with repeat coef """
        random.seed(0)
        reserve = []
        for i in cof_list_all:
            if not cof_list_all:
                reserve.append(i)

        for cof_list in cof_list_all:
            if not cof_list:
                pass
            else:
                if reserve:
                    candi = []
                    for j in cof_list:

                        con = any([[False, True][j in cof_list_all[k]] for k in reserve])
                        if not con:
                            candi.append(j)
                    if any(candi):
                        a = random.choice(candi)
                        reserve.append(a)
                    else:
                        pass
                else:
                    a = random.choice(cof_list)
                    reserve.append(a)
                cof_list_t = copy.deepcopy(cof_list)
                for dela in cof_list_t:
                    for cof_list2 in cof_list_all:
                        if dela in cof_list2:
                            cof_list2.remove(dela)
        return reserve

    @staticmethod
    def cov_y(x_, y_):
        cov = np.corrcoef(x_, y_, rowvar=False)
        cov = cov[:, -1][:-1]
        return cov

    def remove_by_y(self, y_):
        corr = self.cov_y(self.data, y_)
        corr = self.feature_fold(self, corr)
        lcount = self.list_count
        fea_all = []
        score = name_to_name(corr, search=lcount, search_which=0, return_which=(1,), two_layer=True)
        for score_i, list_i in zip(score, lcount):
            indexs = np.argmax(score_i)
            feature_index = list_i[indexs]
            fea_all.append(feature_index)

        fea_all = sorted(list(set(fea_all)))
        return fea_all


if __name__ == "__main__":
    x, y = load_boston(return_X_y=True)
    co = Corr(threshold=0.7)
    c = co.count_cof(np.corrcoef(x, rowvar=False))[1]

    # pandas as pd
    # import matplotlib.pyplot as plt
    # import numpy as np
    #

    #
    # x = np.linspace(0, 1, 1000)
    # y = np.sin(10 * np.pi * x) + x
    # mine = MINE(alpha=0.6, c=15)
    # mine.compute_score(x, y)
    #
    # print("Without noise:")
    # print("MIC", mine.mic())
    # print()
    #
    # np.random.seed(0)
    # y += np.random.uniform(-1, 1, x.shape[0])  # add some noise
    # mine.compute_score(x, y)
    #
    #
    # def MIC_matirx(dataframe, mine):
    #
    #     datamnist = np.array(dataframe)
    #     n = len(datamnist[0, :])
    #     result = np.zeros([n, n])
    #
    #     for i in range(n):
    #         for j in range(n):
    #             mine.compute_score(datamnist[:, i], datamnist[:, j])
    #             result[i, j] = mine.mic()
    #             result[j, i] = mine.mic()
    #     RT = result
    #     return RT
    #
    #
    # mine = MINE(alpha=0.6, c=5)
    # data_wine_mic = MIC_matirx(
    #     np.array([[1, 2, 3, 4, 5, 1, 5, 2, 1, 4], [2, 3, 4, 1, 2, 1, 3, 1, 6, 1], [6, 4, 5, 3, 4, 1, 4, 1, 6, 2]]).T,
    #     mine)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:13
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: GNU Lesser General Public License v3.0

"""
Notes:
    this some of this part are a copy from sympy
"""

from __future__ import division

import numbers

import numpy as np
import numpy.core.numeric as numeric
import sympy
from numpy.linalg import matrix_rank
from sklearn.utils import check_array
from sympy import Add, Mul, Pow, Tuple, sympify
from sympy.core.compatibility import reduce, Iterable
from sympy.physics.units import Dimension
from sympy.physics.units.quantities import Quantity


def dim_map():
    """expr to dim function """

    def my_comp(dim):

        if isinstance(dim, Dim):
            if dim.ndim == 1:
                return dim
            else:
                n = dim.shape[0]
                dc = dim[0].copy()
                return dc.__pow__(n)
        else:
            return dim

    def my_quot(dim):
        if isinstance(dim, Dim):
            if dim.ndim == 1:
                return dim
            elif dim.shape[0] == 2:
                if dim.anyisnan():
                    return dnan
                return dless
            else:
                return dim
        else:
            return dim

    def my_diff(dim):
        if isinstance(dim, Dim):
            if dim.ndim == 1:
                return dim
            elif dim.shape[0] == 2:
                return dim[0].copy()

            else:
                return dim
        else:
            return dim

    def my_conv(dim):
        if isinstance(dim, Dim):
            return dim
        else:
            return dim

    def my_flat(dim):
        if isinstance(dim, Dim):
            if dim.ndim == 1:
                return dim
            else:
                return dim[0].copy()
        else:
            return dim

    def my_abs(dim):
        if isinstance(dim, Dim):
            return dim
        else:
            return dim

    def my_sqrt(dim):

        return dim.__pow__(0.5)

    my_self = my_abs

    def my_exp(dim):

        if isinstance(dim, Dim):
            if dim == dless:
                return dless.get_n(dim)
            else:
                return dnan.get_n(dim)
        else:
            return dim

    my_log = my_ln = my_cos = my_sin = my_exp

    my_funcs = {"Abs": my_abs, "exp": my_exp, "ln": my_ln, 'cos': my_cos, 'sin': my_sin, "log": my_log,
                'sqrt': my_sqrt, "MAdd": my_flat, "MMul": my_comp, "MSub": my_diff,
                "MDiv": my_quot, "Self": my_self, "Conv": my_conv}
    return my_funcs


class Dim(numeric.ndarray):
    """Redefine the Dimension of sympy, the default dimension SI system with 7 number.\n
    1.can be constructed by list of number.\n
    2.can be translated from a sympy.physics.unit. \n
        # >>>from sympy.physics.units import N
        # >>>scale,dim = Dim.convert_to_Dim(N)
        #inverse back
    """
    """
    #     self.unit = [str(i) for i in SI._base_units]\n
    #     self.unit_map = {'meter': "m", 'kilogram': "kg", 'second': "s",
    #     'ampere': "A", 'mole': "mol", 'candela': "cd", 'kelvin': "K"}\n
    #     self.dim = ['length', 'mass', 'time', 'current', 'amount_of_substance',
    #     'luminous_intensity', 'temperature']
    """

    def __new__(cls, data):

        assert isinstance(data, (numeric.ndarray, (list, tuple)))
        dtype = np.float16

        arr = numeric.array(data, dtype=dtype, copy=True)

        shape = arr.shape

        ret = numeric.ndarray.__new__(cls, shape, dtype=np.float16,
                                      buffer=arr,
                                      order='c')
        return ret

    def __eq__(self, other):
        if isinstance(other, Dim):
            se = self.copy()
            ot = other.copy()
            if se.ndim == 2:
                se = se[0]
            if ot.ndim == 2:
                ot = ot[0]
            return all(np.equal(se, ot))
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):

        if isinstance(other, Dim) and self != other:
            return dnan.get_n(self.get_n(other))  # ?
        elif isinstance(other, Dim) and self == other:
            return self.get_n(other)

        else:
            return self

    def __sub__(self, other):
        return self.__add__(other)

    def __pow__(self, other):
        return self._eval_power(other)

    def _eval_power(self, other):
        if isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return Dim(np.array(self) * other)
        else:
            return dnan  #

    def __mul__(self, other):
        if isinstance(other, Dim):
            return Dim(np.array(self) + np.array(other))
        else:
            return self

    def __div__(self, other):

        if isinstance(other, Dim):
            return Dim(np.array(self) - np.array(other))
        else:
            return self

    def __rdiv__(self, other):

        if isinstance(other, (numbers.Real, sympy.Rational, sympy.Float)):
            return self.__pow__(-1)
        elif isinstance(other, Dim):
            return Dim(np.array(other) - np.array(self))
        else:
            return dnan  #

    def __abs__(self):
        return self

    def __rpow__(self, other):
        return dnan  #

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    __truediv__ = __div__
    __rtruediv__ = __rdiv__
    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__

    # @property
    def allisnan(self):
        return np.all(np.isnan(self))

    # @property
    def anyisnan(self):
        return np.any(np.isnan(self))

    # @property
    def isfloat(self):
        return np.any(np.modf(self)[0])

    # @property
    def isinteger(self):
        return not np.any(np.modf(self)[0])

    def is_same_base(self, others):
        se = self.copy()
        if others.ndim == 2:
            others = others[0]
        if se.ndim == 2:
            se = se[0]

        if isinstance(others, Dim):
            npself = np.array(se)
            npothers = np.array(others)
            x1 = np.linalg.norm(npself)
            x2 = np.linalg.norm(npothers)

            if others ** x1 == se ** x2:
                return True
            else:
                return False
        else:
            return False

    def get_n(self, others):
        se = self.copy()

        if others.ndim == 2:
            n = others.shape[0]
        else:
            n = 1

        if self.ndim == 2:
            m = self.shape[0]
        else:
            m = 1

        if m > 1 and n > 1:
            return dnan  #
        elif n == 1:
            return se
        elif m == 1 and n > 1:
            return Dim(np.array([se] * n))
        else:
            return dnan  #

    @staticmethod
    def _get_conversion_matrix_for_expr(expr, target_units, unit_system):
        """depend on sympy 1.5-1.6!!!"""
        from sympy import Matrix

        dimension_system = unit_system.get_dimension_system()

        expr_dim = Dimension(unit_system.get_dimensional_expr(expr))
        dim_dependencies = dimension_system.get_dimensional_dependencies(expr_dim, mark_dimensionless=True)
        target_dims = [Dimension(unit_system.get_dimensional_expr(x)) for x in target_units]
        canon_dim_units = [i for x in target_dims for i in
                           dimension_system.get_dimensional_dependencies(x, mark_dimensionless=True)]
        canon_expr_units = {i for i in dim_dependencies}

        if not canon_expr_units.issubset(set(canon_dim_units)):
            raise TypeError("There is an invalid character in '%s'" % expr,
                            "the expr must be sympy.physics.unit or number")

        seen = set([])
        canon_dim_units = [i for i in canon_dim_units if not (i in seen or seen.add(i))]

        camat = Matrix(
            [[dimension_system.get_dimensional_dependencies(i, mark_dimensionless=True).get(j, 0) for i in target_dims]
             for j in canon_dim_units])
        exprmat = Matrix([dim_dependencies.get(k, 0) for k in canon_dim_units])

        res_exponents = camat.solve_least_squares(exprmat, method=None)

        return res_exponents, canon_dim_units

    @classmethod
    def convert_to(cls, expr, target_units=None, unit_system="SI"):
        """depend on sympy 1.5-1.6!!!"""
        from sympy.physics.units import UnitSystem
        unit_system = UnitSystem.get_unit_system(unit_system)
        if not target_units:
            target_units = unit_system._base_units

        if not isinstance(target_units, (Iterable, Tuple)):
            target_units = [target_units]

        if isinstance(expr, Add):
            raise TypeError("can not be add")

        expr = sympify(expr)

        if not isinstance(expr, Quantity) and expr.has(Quantity):
            expr = expr.replace(lambda x: isinstance(x, Quantity), lambda x: x.convert_to(target_units, unit_system))

        def get_total_scale_factor(expr0):
            if isinstance(expr0, Mul):
                return reduce(lambda x, y: x * y, [get_total_scale_factor(i) for i in expr0.args])
            elif isinstance(expr0, Pow):
                return get_total_scale_factor(expr0.base) ** expr0.exp
            elif isinstance(expr0, Quantity):
                return unit_system.get_quantity_scale_factor(expr0)
            return expr0

        depmat, canon_dim_units = cls._get_conversion_matrix_for_expr(expr, target_units, unit_system)
        if depmat is None:
            raise TypeError("There is an invalid character in '%s'" % expr,
                            "the expr must be sympy.physics.unit or number")

        expr_scale_factor = get_total_scale_factor(expr)
        dim_dict = {}
        for u, p in zip(target_units, depmat):
            expr_scale_factor /= get_total_scale_factor(u) ** p
            dim_dict["%s" % u] = p

        d = cls(np.array(list(dim_dict.values())))
        d.dim = canon_dim_units
        d.unit = target_units
        return expr_scale_factor, d

    @classmethod
    def convert_to_Dim(cls, u, target_units=None, unit_system="SI"):
        """
        depend on sympy 1.5-1.6!!!
        Parameters
        ----------
        u: sympy.physics.unit or Expr of sympy.physics.unit
            unit.
        target_units: None or list of sympy.physics.unit
            if None, the target_units is 7 SI units
        unit_system: str
            default is unit_system="SI"
            
        Returns
        -------
        Expr
        """
        if isinstance(u, Dim):
            return 1, u
        else:
            expr_scale_factor, d = cls.convert_to(u, target_units=target_units, unit_system=unit_system)
            return expr_scale_factor, d

    @classmethod
    def convert_xi(cls, xi, ui, target_units=None, unit_system="SI"):
        """
        depend on sympy 1.5-1.6!!!
        Quick method. translate xi and ui to standard system.
        Parameters
        
        ----------
        xi: np.ndarray
            xi
        ui: sympy.physics.unit or Expr of sympy.physics.unit
            unit
        target_units: None or list of sympy.physics.unit
            if None, the target_units is 7 SI units
        unit_system: str
            default is unit_system="SI"

        Returns
        -------
        xi: 
            np.ndarray
        
        expr: 
            Expr
        """
        expr_scale_factor, d = cls.convert_to_Dim(ui, target_units=target_units, unit_system=unit_system)
        xi = expr_scale_factor * xi
        if isinstance(xi, np.ndarray):
            xi = xi.astype(np.float32)
        return xi, d

    @classmethod
    def convert_x(cls, x, u, target_units=None, unit_system="SI"):
        """
        depend on sympy 1.5-1.6!!!
         Quick method. translate x and u to standard system.
         
         Parameters
         ----------
         x: np.ndarray or list of ndarray,list of float,list of int
             x
         u: list of sympy.physics.unit or Expr of sympy.physics.unit
             units
         target_units: None or list of sympy.physics.unit
             if None, the target_units is 7 SI units
         unit_system: str
             default is unit_system="SI"

         Returns
         -------
         x: 
             np.ndarray
         
         expr: 
             Expr
         
         """
        if isinstance(x, list):
            pass
        elif isinstance(x, np.ndarray):
            x = x.T
        else:
            raise TypeError("values must be list or np.array")
        x_and_d = [cls.convert_xi(xi, x_ui, target_units, unit_system) for xi, x_ui in zip(x, u)]
        x = np.array([xi for xi, d in x_and_d]).T
        x_dim = [d for xi, d in x_and_d]
        return x, x_dim

    @classmethod
    def inverse_convert(cls, dim, scale=1, target_units=None, unit_system="SI"):
        """
        depend on sympy 1.5-1.6!!!
        Quick method. Translate ui to other unit.
        Parameters
        
        ----------
        dim: Dim
            dim
        scale: float
            scale generated before.
        target_units: None or list of sympy.physics.unit
            if None, the target_units is 7 SI units
        unit_system: str
            default is unit_system="SI"

        Returns
        -------
        scale: 
            float
        
        expr: 
            Expr
        """
        from sympy.physics.units import UnitSystem
        unit_system = UnitSystem.get_unit_system(unit_system)
        if not target_units:
            target_units = unit_system._base_units

        if not isinstance(target_units, (Iterable, Tuple)):
            target_units = [target_units]

        def get_total_scale_factor(expr):
            if isinstance(expr, Mul):
                return reduce(lambda x, y: x * y, [get_total_scale_factor(i) for i in expr.args])
            elif isinstance(expr, Pow):
                return get_total_scale_factor(expr.base) ** expr.exp
            elif isinstance(expr, Quantity):
                return unit_system.get_quantity_scale_factor(expr)
            return expr

        sc = scale * Mul.fromiter(1 / get_total_scale_factor(u) ** p for u, p in zip(target_units, dim))
        sc = sc * Mul.fromiter(1 / get_total_scale_factor(u) ** p for u, p in zip(unit_system._base_units, dim))
        tar = Mul.fromiter((1 / get_total_scale_factor(u) * u) ** p for u, p in zip(target_units, dim))
        bas = Mul.fromiter((get_total_scale_factor(u)) ** p for u, p in zip(unit_system._base_units, dim))
        return sc, scale * tar * bas

    @classmethod
    def inverse_convert_xi(cls, xi, dim, scale=1, target_units=None, unit_system="SI"):
        """
        depend on sympy 1.5-1.6!!!
        Quick method. Translate xi, dim to other unit.
        Parameters
        
        ----------
        xi:np.ndarray
            xi
        dim: Dim
            dim
        scale: float
            if xi is have been scaled, the scale is 1.
        target_units: None or list of sympy.physics.unit
            if None, the target_units is 7 SI units
        unit_system: str
            default is unit_system="SI"

        Returns
        -------
        scale: 
            float
        expr: 
            Expr
        """
        expr_scale_factor, d = cls.inverse_convert(dim, scale=scale, target_units=target_units, unit_system=unit_system)
        xi = expr_scale_factor * xi
        if isinstance(xi, np.ndarray):
            xi = xi.astype(np.float32)
        return expr_scale_factor * xi, d


def check_dimension(x, y=None):
    """
    check the consistency of dimension.
    
    Parameters
    ----------
    x: container
        dim of x
    y: Dim
        dim of y

    Returns
    -------
    bool
    """
    if y is not None:
        x.append(y)
    x = np.array(x).T
    x = check_array(x, ensure_2d=True)
    assert isinstance(x, np.ndarray)
    x = x.astype(np.float64)
    det = matrix_rank(x)
    che = []
    for i in range(x.shape[1]):
        x_new = np.delete(x, i, 1)
        det2 = matrix_rank(x_new)
        che.append(det - det2)
    sum(che)

    if sum(che) == 0:
        return True
    else:
        return False


dnan = Dim(np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]))
dless = Dim(np.array([0, 0, 0, 0, 0, 0, 0]))

if __name__ == "__main__":
    a = [Dim([1, 2, 3, 4, 5, 6, 7])] * 100
    b = [Dim([2, 2, 3, 4, 5, 6, 7])] * 100

    dl = dless
    dn = dnan

    c = Dim([1, 2, 3, 4, 5, 6, 7])

    t = zip(a, b)


    def func(k):
        ss = k[1] * k[0]
        return ss



# -*- coding: utf-8 -*-

# @Time    : 2019/10/13 21:55
# @Email   : 986798607@qq.ele_ratio
# @Software: PyCharm
# @License: BSD 3-Clause
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 28 15:24:10 2018

@author: ww
"""

import numpy as np
import sklearn.utils
from scipy import stats
from sklearn.utils import check_array

from featurebox.tools.tool import parallelize

print('\t\tFor grid building\nsample:\nspace=searchspace(li1,li2,li3)\nNote:parameters no more than 5 ')
print(
    '\n\t\tFor ego,kg,maxp\nSample:\nresults=egosearch(X=?,searchspace=?,number=?,regclf=RandomForestRegressor(),'
    'rankway="ego"or"kg"or"maxp"or"No",meanstd=False)')
print(
    'return:\nresult is 2 dimentions array\n1st column = sequence number,\n2nd part = your searchspace,\n3rd part = '
    'mean,std,ego,kg,maxp,sequentially')


def search_space(*arg):
    meshes = np.meshgrid(*arg)
    meshes = [_.ravel() for _ in meshes]
    meshes = np.array(meshes).T
    return meshes


class Ego:
    def __init__(self, searchspace, X, y, number, regclf, n_jobs=2):
        self.n_jobs = n_jobs
        check_array(X, ensure_2d=True, force_all_finite=True)
        check_array(y, ensure_2d=True, force_all_finite=True)
        check_array(searchspace, ensure_2d=True, force_all_finite=True)
        assert X.shape[1] == X.searchspace[1]
        self.searchspace = searchspace
        self.X = X
        self.y = y
        self.regclf = regclf

        self.meanandstd_all = []
        self.predict_y_all = []
        self.number = number

    def Fit(self):
        x = self.X
        y = self.y
        njobs = self.n_jobs
        searchspace0 = self.searchspace
        regclf0 = self.regclf

        def fit_parllize(random_state):
            data_train, y_train = sklearn.utils.resample(x, y, n_samples=None, replace=True,
                                                         random_state=random_state)
            regclf0.fit(data_train, y_train)
            predict_data = regclf0.predict(searchspace0)
            predict_data.ravel()

        predict_dataj = parallelize(n_jobs=njobs, func=fit_parllize, iterable=range(self.number))

        return np.array(predict_dataj)

    @staticmethod
    def meanandstd(predict_dataj):
        mean = np.mean(predict_dataj, axis=1)
        std = np.std(predict_dataj, axis=1)
        data_predict = np.column_stack((mean, std))
        print(data_predict.shape)
        return data_predict

    @staticmethod
    def CalculateEi(y, meanstd0):
        ego = (meanstd0[:, 0] - max(y)) / (meanstd0[:, 1])
        ei_ego = meanstd0[:, 1] * ego * stats.norm.cdf(ego) + meanstd0[:, 1] * stats.norm.pdf(ego)
        kg = (meanstd0[:, 0] - max(max(meanstd0[:, 0]), max(y))) / (meanstd0[:, 1])
        ei_kg = meanstd0[:, 1] * kg * stats.norm.cdf(kg) + meanstd0[:, 1] * stats.norm.pdf(kg)
        max_P = stats.norm.cdf(ego, loc=meanstd0[:, 0], scale=meanstd0[:, 1])
        ei = np.column_stack((meanstd0, ei_ego, ei_kg, max_P))
        print('ego is done')
        return ei

    def egosearch(self, rankway="ego", meanstd=False):
        y = self.y
        searchspace0 = self.searchspace
        if rankway not in ['ego', 'kg', 'maxp', 'no', 'No']:
            print('Don\'t kidding me,checking rankway=what?\a')
        else:
            if not meanstd:
                predict_data = self.Fit()
                meanstd = self.meanandstd(predict_data)
            else:
                pass

            result = self.CalculateEi(y, meanstd)
            bianhao = np.arange(0, len(result))
            result1 = np.column_stack((bianhao, searchspace0, result))
            if rankway == "No" or "no":
                pass
            if rankway == "ego":
                egopaixu = np.argsort(result1[:, -3])
                result1 = result1[egopaixu]
            elif rankway == "kg":
                kgpaixu = np.argsort(result1[:, -2])
                result1 = result1[kgpaixu]
            elif rankway == "maxp":
                max_paixu = np.argsort(result1[:, -1])
                result1 = result1[max_paixu]
            return result1

    def Rank(self, rankway="ego", meanstd=False):
        return self.egosearch(self, rankway, meanstd=meanstd)



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/28 16:26
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
this is a description
"""

from __future__ import division

from abc import ABC
from itertools import permutations

import numpy as np
from sklearn.utils import check_array
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted

from featurebox.featurizers.base import BaseFeaturizer


class PolyFeaturizer(BaseFeaturizer, ABC):

    def __init__(self, *, degree=2, n_jobs=-1, on_errors='raise', return_type='any'):
        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

        if isinstance(degree, (float, int)):
            degree = [degree, ]
        if 0 not in degree:
            degree.append(0)
        self.degree = degree

    @staticmethod
    def _combinations(n_features, degree):
        assert len(degree) ** n_features <= 1e6, "too much degree to calculate, plese depress the degree"
        return permutations(degree, n_features)

    def fit(self, X, **kwargs):
        n_samples, n_features = check_array(X, accept_sparse=True).shape
        self.n_input_features_ = n_features
        return self

    def transform(self, X, y=None, **kwargs):
        check_is_fitted(self, ['n_input_features_'])

        X = check_array(X, dtype=FLOAT_DTYPES, accept_sparse='csc')
        n_samples, n_features = X.shape

        if n_features != self.n_input_features_:
            raise ValueError("X shape does not match training shape")

        combinations = self._combinations(n_features, self.degree, )
        columns = []
        for comb in combinations:
            out_col = 1
            for col_idx, combi in enumerate(comb):
                out_col = X[:, col_idx] ** combi * out_col
            columns.append(out_col)
        XP = np.vstack(columns).T
        return XP

    def feature_labels(self, input_features=None):

        check_is_fitted(self, 'n_input_features_')
        if input_features is None:
            input_features = ['x%d' % i for i in range(self.n_input_features_)]
        combinations = self._combinations(self.n_input_features_, self.degree)
        feature_names = []
        for rows in combinations:
            names = "*".join(["{}^{}".format(feature, row) for row, feature in zip(rows, input_features) if row != 0])

            feature_names.append(names)
        return feature_names



import unittest

import numpy
from featurebox.symbol.functions.dimfunc import dless, Dim

from featurebox.symbol.base import CalculatePrecisionSet
from featurebox.symbol.base import SymbolSet
from featurebox.symbol.base import SymbolTree


class MyTestbase(unittest.TestCase):

    def setUp(self):
        self.SymbolTree = SymbolTree
        self.pset = SymbolSet()

        from sklearn.datasets import load_boston

        data = load_boston()
        x = data["data"]
        y = data["target"]
        # No = Normalizer()
        # y=y/max(y)
        # x = No.fit_transform(x)
        self.x = x
        self.y = y
        # self.pset.add_features(x, y, )
        self.pset.add_features(x, y, x_group=[[1, 2], [4, 5]])
        self.pset.add_constants([6, 3, 4], c_dim=[dless, dless, dless], c_prob=None)
        self.pset.add_operations(power_categories=(2, 3, 0.5),
                                 categories=("Add", "Mul", "Self", "Abs"),
                                 self_categories=None)

        from sklearn.metrics import r2_score, mean_squared_error
        self.cp = CalculatePrecisionSet(self.pset, scoring=[r2_score, mean_squared_error],
                                        score_pen=[1, -1],
                                        filter_warning=True)

    def test_pset_passed_to_cpset_will_change(self):
        cp = CalculatePrecisionSet(self.pset)
        self.assertNotEqual(cp, self.cp)

    def test_tree_gengrow_repr_and_str_different(self):
        from numpy import random
        random.seed(1)
        sl = SymbolTree.genGrow(self.pset, 3, 4)
        print(sl)
        # self.assertNotEqual(repr(sl), str(sl))

    def test_add_tree_back(self):
        from numpy import random
        random.seed(1)
        sl = SymbolTree.genGrow(self.pset, 3, 4)
        self.pset.add_tree_to_features(sl)

    #
    def test_barch_tree(self):
        from numpy import random
        random.seed(1)
        for i in range(10):

            sl = SymbolTree.genGrow(self.pset, 3, 4)
            cpsl = self.cp.calculate_detail(sl)
            self.assertIsNotNone(cpsl.y_dim)
            self.assertIsNotNone(cpsl.expr)
            self.assertIsNone(cpsl.p_name)
            if cpsl.pre_y is not None:
                self.assertIsInstance(cpsl.pre_y, numpy.ndarray)
                self.assertEqual(cpsl.pre_y.shape, self.y.shape)
                print(cpsl.coef_pre_y[:3])
                print(cpsl.pre_y[:3])
                print(cpsl.coef_score)
                print(cpsl.coef_expr)
                print(cpsl.pure_expr)

    def test_depart_tree(self):
        from numpy import random
        random.seed(1)
        for i in range(10):

            sl = SymbolTree.genGrow(self.pset, 5, 6)
            sl_departs = sl.depart()
            for i in sl_departs:
                cpsl = self.cp.calculate_simple(i)
                self.assertIsNotNone(cpsl.y_dim)
                self.assertIsNotNone(cpsl.expr)
                self.assertIsNone(cpsl.p_name)


if __name__ == '__main__':
    unittest.main()



from statistics import mode, mean

import torch
import torch.nn.functional as F


class SaveValues():
    def __init__(self, m):
        # register a hook to save values of activations and gradients
        self.activations = None
        self.gradients = None
        self.forward_hook = m.register_forward_hook(self.hook_fn_act)
        self.backward_hook = m.register_backward_hook(self.hook_fn_grad)

    def hook_fn_act(self, module, input, output):
        self.activations = output

    def hook_fn_grad(self, module, grad_input, grad_output):
        self.gradients = grad_output[0]

    def remove(self):
        self.forward_hook.remove()
        self.backward_hook.remove()


class CAM(object):
    """ Class Activation Mapping """

    def __init__(self, model, target_layer):
        """
        Args:
            model: a base model to get CAM which have global pooling and fully connected layer.
            target_layer: conv_layer before Global Average Pooling
        """

        self.model = model
        self.target_layer = target_layer

        # save values of activations and gradients in target_layer
        self.values = SaveValues(self.target_layer)

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
        Return:
            heatmap: class activation mappings of the predicted class
        """

        # object classification
        score = self.model(x)

        prob = F.softmax(score, dim=1)

        if idx is None:
            prob, idx = torch.max(prob, dim=1)
            idx = idx.item()
            prob = prob.item()
            print("predicted class ids {}\t probability {}".format(idx, prob))

        # cam can be calculated from the weights of linear layer and activations
        weight_fc = list(
            self.model._modules.get('fc').parameters())[0].to('cpu').data

        cam = self.getCAM(self.values, weight_fc, idx)

        return cam, idx

    def __call__(self, x):
        return self.forward(x)

    def getCAM(self, values, weight_fc, idx):
        '''
        values: the activations and gradients of target_layer
            activations: feature map before GAP.  shape => (1, C, H, W)
        weight_fc: the weight of fully connected layer.  shape => (num_classes, C)
        idx: predicted class id
        cam: class activation map.  shape => (1, num_classes, H, W)
        '''

        cam = F.conv2d(values.activations, weight=weight_fc[:, :, None, None])
        _, _, h, w = cam.shape

        # class activation mapping only for the predicted class
        # cam is normalized with min-max.
        cam = cam[:, idx, :, :]
        cam -= torch.min(cam)
        cam /= torch.max(cam)
        cam = cam.view(1, 1, h, w)

        return cam.data


class GradCAM(CAM):
    """ Grad CAM """

    def __init__(self, model, target_layer):
        super().__init__(model, target_layer)

        """
        Args:
            model: a base model to get CAM, which need not have global pooling and fully connected layer.
            target_layer: conv_layer you want to visualize
        """

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
            idx: ground truth index => (1, C)
        Return:
            heatmap: class activation mappings of the predicted class
        """

        # anomaly detection
        score = self.model(x)

        prob = F.softmax(score, dim=1)

        if idx is None:
            prob, idx = torch.max(prob, dim=1)
            idx = idx.item()
            prob = prob.item()
            print("predicted class ids {}\t probability {}".format(idx, prob))

        # caluculate cam of the predicted class
        cam = self.getGradCAM(self.values, score, idx)

        return cam, idx

    def __call__(self, x):
        return self.forward(x)

    def getGradCAM(self, values, score, idx):
        '''
        values: the activations and gradients of target_layer
            activations: feature map before GAP.  shape => (1, C, H, W)
        score: the output of the model before softmax
        idx: predicted class id
        cam: class activation map.  shape=> (1, 1, H, W)
        '''

        self.model.zero_grad()

        score[0, idx].backward(retain_graph=True)

        activations = values.activations
        gradients = values.gradients
        n, c, _, _ = gradients.shape

        alpha = gradients.view(n, c, -1).mean(2)
        alpha = alpha.view(n, c, 1, 1)

        # shape => (1, 1, H', W')
        cam = (alpha * activations).sum(dim=1, keepdim=True)
        cam = F.relu(cam)
        cam -= torch.min(cam)
        cam /= torch.max(cam)

        return cam.data


class GradCRM(CAM):
    """ Grad CAM """

    def __init__(self, model, target_layer):
        super().__init__(model, target_layer)

        """
        Args:
            model: a base model to get CAM, which need not have global pooling and fully connected layer.
            target_layer: conv_layer you want to visualize
        """

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
            idx: ground truth index => (1, C)
        Return:
            heatmap: class activation mappings of the predicted class
        """

        # anomaly detection
        score = self.model(x)

        # caluculate cam of the predicted class
        cam = self.getGradCAM(self.values, score, idx)

        return cam, idx

    def __call__(self, x, idx):
        return self.forward(x, idx)

    def getGradCAM(self, values, score, idx):
        '''
        values: the activations and gradients of target_layer
            activations: feature map before GAP.  shape => (1, C, H, W)
        score: the output of the model before softmax
        idx: predicted class id
        cam: class activation map.  shape=> (1, 1, H, W)
        '''

        self.model.zero_grad()

        score.backward(retain_graph=True)

        activations = values.activations
        gradients = values.gradients
        n, c, _, _ = gradients.shape

        alpha = gradients.view(n, c, -1).mean(2)
        alpha = alpha.view(n, c, 1, 1)

        # shape => (1, 1, H', W')
        cam = (alpha * activations).sum(dim=1, keepdim=True)
        cam = F.relu(cam)
        cam -= torch.min(cam)
        cam /= torch.max(cam)

        return cam.data


class GradCRMpp(CAM):
    """ Grad CAM plus plus """

    def __init__(self, model, target_layer):
        super().__init__(model, target_layer)
        """
        Args:
            model: a base model
            target_layer: conv_layer you want to visualize
        """

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
        Return:
            heatmap: class activation mappings of predicted classes
        """

        # object classification
        score = self.model(x)

        # caluculate cam of the predicted class
        cam = self.getGradCAMpp(self.values, score, idx)

        return cam, idx

    def __call__(self, x, idx):
        return self.forward(x, idx)

    def getGradCAMpp(self, values, score, idx):
        '''
        values: the activations and gradients of target_layer
            activations: feature map before GAP.  shape => (1, C, H, W)
        score: the output of the model before softmax. shape => (1, n_classes)
        idx: predicted class id
        cam: class activation map.  shape=> (1, 1, H, W)
        '''

        self.model.zero_grad()

        score[idx].backward(retain_graph=True)

        activations = values.activations
        gradients = values.gradients
        n, c, _, _ = gradients.shape

        # calculate alpha
        numerator = gradients.pow(2)
        denominator = 2 * gradients.pow(2)
        ag = activations * gradients.pow(3)
        denominator += ag.view(n, c, -1).sum(-1, keepdim=True).view(n, c, 1, 1)
        denominator = torch.where(
            denominator != 0.0, denominator, torch.ones_like(denominator))
        alpha = numerator / (denominator + 1e-7)

        relu_grad = F.relu(score[0, idx].exp() * gradients)
        weights = (alpha * relu_grad).view(n, c, -1).sum(-1).view(n, c, 1, 1)

        # shape => (1, 1, H', W')
        cam = (weights * activations).sum(1, keepdim=True)
        cam = F.relu(cam)
        cam -= torch.min(cam)
        cam /= torch.max(cam)

        return cam.data


class GradCAMpp(CAM):
    """ Grad CAM plus plus """

    def __init__(self, model, target_layer):
        super().__init__(model, target_layer)
        """
        Args:
            model: a base model
            target_layer: conv_layer you want to visualize
        """

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
        Return:
            heatmap: class activation mappings of predicted classes
        """

        # object classification
        score = self.model(x)

        prob = F.softmax(score, dim=1)

        if idx is None:
            prob, idx = torch.max(prob, dim=1)
            idx = idx.item()
            prob = prob.item()
            print("predicted class ids {}\t probability {}".format(idx, prob))

        # caluculate cam of the predicted class
        cam = self.getGradCAMpp(self.values, score, idx)

        return cam, idx

    def __call__(self, x):
        return self.forward(x)

    def getGradCAMpp(self, values, score, idx):
        '''
        values: the activations and gradients of target_layer
            activations: feature map before GAP.  shape => (1, C, H, W)
        score: the output of the model before softmax. shape => (1, n_classes)
        idx: predicted class id
        cam: class activation map.  shape=> (1, 1, H, W)
        '''

        self.model.zero_grad()

        score[0, idx].backward(retain_graph=True)

        activations = values.activations
        gradients = values.gradients
        n, c, _, _ = gradients.shape

        # calculate alpha
        numerator = gradients.pow(2)
        denominator = 2 * gradients.pow(2)
        ag = activations * gradients.pow(3)
        denominator += ag.view(n, c, -1).sum(-1, keepdim=True).view(n, c, 1, 1)
        denominator = torch.where(
            denominator != 0.0, denominator, torch.ones_like(denominator))
        alpha = numerator / (denominator + 1e-7)

        relu_grad = F.relu(score[0, idx].exp() * gradients)
        weights = (alpha * relu_grad).view(n, c, -1).sum(-1).view(n, c, 1, 1)

        # shape => (1, 1, H', W')
        cam = (weights * activations).sum(1, keepdim=True)
        cam = F.relu(cam)
        cam -= torch.min(cam)
        cam /= torch.max(cam)

        return cam.data


class SmoothGradCAMpp(CAM):
    """ Smooth Grad CAM plus plus """

    def __init__(self, model, target_layer, n_samples=25, stdev_spread=0.15):
        super().__init__(model, target_layer)
        """
        Args:
            model: a base model
            target_layer: conv_layer you want to visualize
            n_sample: the number of samples
            stdev_spread: standard deviationß
        """

        self.n_samples = n_samples
        self.stdev_spread = stdev_spread

    def forward(self, x, idx=None):
        """
        Args:
            x: input image. shape =>(1, 3, H, W)
        Return:
            heatmap: class activation mappings of predicted classes
        """

        stdev = self.stdev_spread / (x.max() - x.min())
        std_tensor = torch.ones_like(x) * stdev

        indices = []
        probs = []

        for i in range(self.n_samples):
            self.model.zero_grad()

            x_with_noise = torch.normal(mean=x, std=std_tensor)
            x_with_noise.requires_grad_()

            score = self.model(x_with_noise)

            prob = F.softmax(score, dim=1)

            if idx is None:
                prob, idx = torch.max(prob, dim=1)
                idx = idx.item()
                probs.append(prob.item())

            indices.append(idx)

            score[0, idx].backward(retain_graph=True)

            activations = self.values.activations
            gradients = self.values.gradients
            n, c, _, _ = gradients.shape

            # calculate alpha
            numerator = gradients.pow(2)
            denominator = 2 * gradients.pow(2)
            ag = activations * gradients.pow(3)
            denominator += \
                ag.view(n, c, -1).sum(-1, keepdim=True).view(n, c, 1, 1)
            denominator = torch.where(
                denominator != 0.0, denominator, torch.ones_like(denominator))
            alpha = numerator / (denominator + 1e-7)

            relu_grad = F.relu(score[0, idx].exp() * gradients)
            weights = \
                (alpha * relu_grad).view(n, c, -1).sum(-1).view(n, c, 1, 1)

            # shape => (1, 1, H', W')
            cam = (weights * activations).sum(1, keepdim=True)
            cam = F.relu(cam)
            cam -= torch.min(cam)
            cam /= torch.max(cam)

            if i == 0:
                total_cams = cam.clone()
            else:
                total_cams += cam

        total_cams /= self.n_samples
        idx = mode(indices)
        prob = mean(probs)

        print("predicted class ids {}\t probability {}".format(idx, prob))

        return total_cams.data, idx

    def __call__(self, x):
        return self.forward(x)

# if __name__=="__main__":
#
#     import torch
#     import torch.nn. functional as F
#     from matplotlib.pyplot import imshow
#     from torchvision import models, transforms
#     import os
#
#     os.chdir(r"C:\Users\Administrator\Desktop\fig")
#     from  skimage import io
#     image = io.imread('dog.png')
#
#     imshow(image)
#
#     preprocess = transforms.Compose([
#     transforms.ToTensor(),
#     ])
#     # convert image to tensor
#     tensor = preprocess(image)
#     tensor = normalize(tensor)
#     # reshape 4D tensor (N, C, H, W)
#     tensor = tensor.unsqueeze(0)
#
#     model = models.resnet18(pretrained=True)
#     model.eval()
#     print(model)
#
#     # In [7]:
#     # # the target layer you want to visualize
#     target_layer = model.layer4[1].conv2
#     #
#     # # wrapper for class activation mapping. Choose one of the following.
#     # wrapped_model = CAM(model, target_layer)
#     # wrapped_model = GradCAM(model, target_layer)
#     wrapped_model = GradCAMpp(model, target_layer)
#     # wrapped_model = SmoothGradCAMpp(model, target_layer, n_samples=5, stdev_spread=0.15)
#     # In [8]:
#     cam, idx = wrapped_model(tensor)
#
#     # # visualize only cam
#     imshow(cam.squeeze().numpy(), alpha=0.5, cmap='jet')
#
#     # In [11]:
#     # # reverse normalization for display
#     img = normalize.inverse_transform(tensor)
#     # In [12]:
#     heatmap = visualize(img, cam)
#     # In [13]:
#     # # save image
#     ##save_image(heatmap, './sample/{}_cam.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
#     # # save_image(heatmap, './sample/{}_gradcam.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
#     # # save_image(heatmap, './sample/{}_gradcampp.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
#     # save_image(heatmap, './sample/{}_smoothgradcampp.png'.format(idx2label[idx]).replace(" ", "_").replace(",", ""))
#     # In [14]:
#     # # or visualize on Jupyter
#     hm = (heatmap.squeeze().numpy().transpose(1, 2, 0))
#     imshow(hm)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# @Time    : 2019/11/12 15:13
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: GNU Lesser General Public License v3.0

"""
Notes:
    this part are one copy from deap,
    change the random to numpy.random
"""

import copy
import operator
import sys
from collections import Counter
from functools import wraps
from inspect import isclass
from operator import attrgetter

import numpy as np
from deap.tools import Statistics, MultiStatistics
from numpy import random

from featurebox.symbol.calculation.scores import score_dim


######################################
# Generate                         #
######################################


def checkss(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)

        pset = kwargs["pset"]
        for i in result[0].bot():
            assert i in pset.dispose
        for i in result[0].top():
            assert i in pset.primitives + pset.terminals_and_constants

        return result

    return wrapper


def checks_number(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if "force_number" in kwargs:
            fs = kwargs["force_number"]
        else:
            fs = False
        if fs and not len(result):
            raise TypeError("The DimForceLoop keep the dim of the select offspring,\n"
                            "But the select number from last population are zero at the current dim_type limitation.\n"
                            "Please change the dim_type or change the DimForceLoop method to others")

        return result

    return wrapper


def generate(pset, min_, max_, condition, personal_map=False, *kwargs):
    """

    Parameters
    ----------
    pset: SymbolSet
        pset
    min_: int
        Minimum height of the produced trees.
    max_: int
        Maximum Height of the produced trees.
    condition: collections.Callable
        The condition is a function that takes two arguments,
        the height of the tree to build and the current
        depth in the tree.
    kwargs: None
        placeholder for future
    personal_map:bool
        premap

    Returns
    -------

    """
    _ = kwargs
    type_ = object
    expr = []
    height = random.randint(min_, max_)
    stack = [(0, type_)]
    while len(stack) != 0:
        depth, type_ = stack.pop()
        if condition(height, depth):
            try:
                if personal_map:
                    p_t = pset.premap.get_ind_value(expr, pset)
                else:
                    p_t = pset.prob_ter_con_list
                if p_t is None:
                    p_t = pset.prob_ter_con_list

                term = pset.terminals_and_constants[random.choice(len(pset.terminals_and_constants),
                                                                  p=p_t)]

            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The symbol.generate function tried to add "
                                 "a terminalm, but there is "
                                 "none available.").with_traceback(traceback)
            if isclass(term):
                term = term()
            expr.append(term)
        else:
            try:
                prim = pset.primitives[random.choice(len(pset.primitives), p=pset.prob_pri_list)]
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The symbol.generate function tried to add "
                                 "a primitive', but there is "
                                 "none available.").with_traceback(traceback)
            expr.append(prim)
            for arg in reversed(prim.args):
                stack.append((depth + 1, arg))

    dispose = list(random.choice(pset.dispose, len(expr), p=pset.prob_dispose_list))

    if pset.types == 1:
        add_ = [pset.dispose_dict["Self"]]
        dispose = add_*len(expr)

    elif pset.types == 2:
        add_ = list(pset.dispose_dict[i] for i in ["MAdd", "MSub", "MMul", "MDiv"] )
        dispose[0] = random.choice(add_, p=[0.25, 0.25, 0.25, 0.25])
    else:
        add_ = list(pset.dispose_dict[i] for i in ["MAdd", "MMul"] )
        dispose[0] = random.choice(add_, p=[0.5, 0.5])

    re = []
    for i, j in zip(dispose, expr):
        re.extend((i, j))

    return re


def genGrow(pset, min_, max_, personal_map=False, ):
    """Generate an expression where each leaf might have a different depth
    between *min* and *max*.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param personal_map: bool.
    
    :returns: A grown tree with leaves at possibly different depths.

    """

    def condition(height, depth):
        """Expression generation stops when the depth is equal to height
        or when it is randomly determined that a node should be a terminal.
        """
        return depth == height or (depth >= min_ and random.random() < pset.terminalRatio)

    return generate(pset, min_, max_, condition, personal_map=personal_map)


def depart(individual):
    if len(individual) <= 10 or individual.height <= 8:
        return [individual, ]
    else:
        inds = []
        for index in np.arange(2, len(individual) - 4, step=2):
            slice_ = individual.searchSubtree(index)
            ind_new = individual.__class__(individual[slice_])
            if 6 <= len(ind_new) <= 10 or 4 <= ind_new.height <= 8:
                if len(ind_new.ter_site()) >= 2:
                    ind_new[0] = individual[0]
                    inds.append(ind_new)

        return inds


def genFull(pset, min_, max_, personal_map=False):
    """Generate an expression where each leaf has the same depth
    between *min* and *max*.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param personal_map:

    :returns: A full tree with all leaves at the same depth.
    """

    def condition(height, depth):
        """Expression generation stops when the depth is equal to height."""
        return depth == height

    return generate(pset, min_, max_, condition, personal_map=personal_map)


######################################
# crossover                        #
######################################
def cxOnePoint(ind10, ind20):
    """Randomly select crossover point in each individual and exchange each
    subtree with the point as root between each individual.

    :param ind10: First tree participating in the crossover.
    :param ind20: Second tree participating in the crossover.
    :returns: A tuple of two trees.
    """
    ind1 = copy.copy(ind10)
    ind2 = copy.copy(ind20)

    if len(ind1) < 4 or len(ind2) < 4:

        return ind1, ind2
    #
    else:
        index1 = random.choice(np.arange(2, len(ind1) - 1, 2))
        index2 = random.choice(np.arange(2, len(ind2) - 1, 2))
        slice1 = ind1.searchSubtree(index1)
        slice2 = ind2.searchSubtree(index2)

        ind1[slice1], ind2[slice2] = ind2[slice2], ind1[slice1]

    return ind1, ind2


######################################
# limitation                       #
######################################


def staticLimit(key, max_value):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):

            keep_inds = [copy.deepcopy(ind) for ind in args]
            new_inds = list(func(*args, **kwargs))
            for i, ind in enumerate(new_inds):

                if key(ind) == key(keep_inds[i]):
                    pass
                elif key(ind) > max_value:
                    new_inds[i] = keep_inds[random.choice(len(keep_inds))]

            return new_inds

        return wrapper

    return decorator


######################################
# mutate                       #
######################################
# @logg
# @checkss
def mutUniform(individual, expr, pset):
    """Randomly select a point in the tree *individual*, then replace the
    subtree at that point as a root by the expression generated using method
    :func:`expr`.

    :param individual: The tree to be mutated.
    :param expr: A function object that can generate an expression when
                 called.
    :param pset: SymbolSet
    :returns: A tuple of one tree.

    """
    index = random.choice(len(individual))
    if index % 2 == 1:
        index -= 1
    slice_ = individual.searchSubtree(index)

    individual[slice_] = expr(pset=pset)
    return individual,


# @logg
# @checkss
def mutShrink(individual, pset=None):
    """This operator shrinks the *individual* by choosing randomly a branch and
    replacing it with one of the branch's arguments (also randomly chosen).

    :param individual: The tree to be shrinked.
    :param pset: SymbolSet.
    :returns: A tuple of one tree.
    """
    _ = pset
    # We don't want to "shrink" the root
    if len(individual) < 4 or individual.height <= 4:
        return individual,

    index = random.randint(0, len(individual))

    if index % 2 == 1:
        index -= 1
    slice_ = individual.searchSubtree(index)

    ter = [i for i in individual[slice_] if i.arity == 0]
    left = random.choice(ter)
    hat = random.choice(pset.dispose, p=pset.prob_dispose_list)

    del individual[slice_]
    individual.insert(index, left)
    individual.insert(index, hat)

    return individual,


# @logg
# @checkss
def mutNodeReplacementVerbose(individual, pset, personal_map=False):
    """
    # choice terminals_and_constants verbose
    Replaces a randomly chosen primitive from *individual* by a randomly
    chosen primitive with the same number of arguments from the :attr:`pset`
    attribute of the individual.

    :param individual: The normal or typed tree to be mutated.
    :param pset: SymbolSet
    :param personal_map: bool
    :returns: A tuple of one tree.
    """

    if len(individual) < 4:
        return individual,

    if pset.types > 1:
        if random.random() <= 0.8:
            index = random.choice(np.arange(1, len(individual), step=2))
        else:
            index = random.choice(np.arange(0, len(individual), step=2))
    else:
        index = random.choice(np.arange(0, len(individual), step=2))
    node = individual[index]


    if index % 2 == 0:
        for i in pset.dispose:
            assert i.arity == 1
        prims = pset.dispose
        p_d = np.array([pset.prob_dispose[repr(i)] for i in prims], 'float32')
        p_d /= np.sum(p_d)
        a = prims[random.choice(len(prims), p=p_d)]
        individual[index] = a
    else:

        if node.arity == 0:  # Terminal
            if personal_map:
                p_t = pset.premap.get_one_node_value(individual, pset, node=node, site=index)
                if p_t is None:
                    p_t = pset.prob_ter_con_list
            else:
                p_t = pset.prob_ter_con_list

            term = pset.terminals_and_constants[random.choice(len(pset.terminals_and_constants), p=p_t)]

            individual[index] = term
        else:  # Primitive
            prims = [p for p in pset.primitives if p.arity == node.arity]
            p_p = np.array([pset.prob_pri[repr(i)] for i in prims], 'float32')

            p_p /= np.sum(p_p)
            # except:
            a = prims[random.choice(len(prims), p=p_p)]

            individual[index] = a

    return individual,


# @logg
# @checkss
def mutDifferentReplacementVerbose(individual, pset, personal_map=False):
    """
    # choice terminals_and_constants verbose
    Replaces a randomly chosen primitive from *individual* by a randomly
    chosen primitive with the same number of arguments from the :attr:`pset`
    attribute of the individual.
    decrease the probability of same terminals.

    :param individual: The normal or typed tree to be mutated.
    :param pset: SymbolSet
    :param personal_map: bool

    :returns: A tuple of one tree.
    """
    if len(individual) < 4:
        return individual,
    ters = [repr(i) for i in individual.terminals()]
    pset_ters = [repr(i) for i in pset.terminals_and_constants]
    cou = Counter(ters)
    cou_mutil = {i: j for i, j in cou.items() if j >= 2}
    ks = list(cou_mutil.keys())
    nks = list(set(pset_ters) - (set(ks)))
    nks.sort()  # very import for random

    p_nks = np.array([pset.prob_ter_con[i] for i in nks])
    p_nks /= np.sum(p_nks)

    if cou_mutil:
        indexs = []
        for k, v in cou_mutil.items():
            indi = []
            for i in np.arange(1, len(individual), 2):
                if repr(individual[i]) == k:
                    indi.append(i)
            if indi:
                indexs.append(random.choice(indi))

        if personal_map:
            p_nks_new = pset.premap.get_nodes_value(ind=individual, pset=pset, node=None, site=indexs)
            if p_nks_new is not None:
                nks = list(pset.prob_ter_con.keys())
                p_nks = p_nks_new

        if len(indexs) <= len(nks):
            term = random.choice(nks, len(indexs), replace=False, p=p_nks)
        else:
            term = random.choice(nks, len(indexs), replace=True, p=p_nks)

        term_ters = []
        for name in term:
            for i in pset.terminals_and_constants:
                if repr(i) == name:
                    term_ters.append(i)

        for o, n in zip(indexs, term_ters):
            individual[o] = n

    return individual,


######################################
# select                       #
######################################


def selRandom(individuals, k):
    """Select *k* individuals at random from the input *individuals* with
    replacement. The list returned contains references to the input
    *individuals*.

    :param individuals: A list of individuals to select from.
    :param k: The number of individuals to select.
    :returns: A list of selected individuals.

    This function uses the :func:`numpy.random.choice` function
    """
    return [individuals[random.choice(len(individuals))] for _ in range(k)]


def selBest(individuals, k, fit_attr="fitness"):
    """Select the *k* best individuals among the input *individuals*. The
    list returned contains references to the input *individuals*.

    :param individuals: A list of individuals to select from.
    :param k: The number of individuals to select.
    :param fit_attr: The attribute of individuals to use as selection criterion
    :returns: A list containing the k best individuals.
    """
    return sorted(individuals, key=attrgetter(fit_attr), reverse=True)[:k]


def selTournament(individuals, k, tournsize, fit_attr="fitness"):
    """Select the best individual among *tournsize* randomly chosen
    individuals, *k* times. The list returned contains
    references to the input *individuals*.

    :param individuals: A list of individuals to select from.
    :param k: The number of individuals to select.
    :param tournsize: The number of individuals participating in each tournament.
    :param fit_attr: The attribute of individuals to use as selection criterion
    :returns: A list of selected individuals.

    This function uses the :func:`numpy.random.choice` function
    """
    chosen = []
    for i in range(k):
        aspirants = selRandom(individuals, tournsize)
        chosen.append(max(aspirants, key=attrgetter(fit_attr)))
    return chosen


@checks_number
def selKbestDim(pop, K_best=10, dim_type=None, fuzzy=False, fit_attr="fitness", force_number=False):
    """
    Select the individual with dim limitation.
    Parameters
    
    ----------
    pop: SymbolTree
        A list of individuals to select from.
    K_best:int
        The number of individuals to select.
    dim_type:Dim
    fuzzy:bool
        the dim or the dim with same base. such as m,m^2,m^3
    fit_attr:str
        The attribute of individuals to use as selection criterion
    force_number:
        return the number the same with K
    .. note::
        default attr is "fitness"
        
    Returns
    -------
    A list of selected individuals.
    """
    chosen = sorted(pop, key=operator.attrgetter(fit_attr))
    chosen.reverse()

    choice_index = [score_dim(ind.y_dim, dim_type, fuzzy) for ind in chosen]
    add_ind = [chosen[i] for i, j in enumerate(choice_index) if j == 1]

    if K_best is None:
        if len(add_ind) >= 5:
            K_best = round(len(pop) / 10)
        else:
            K_best = 0
    if len(add_ind) >= round(K_best):
        return add_ind[:round(K_best)]
    else:
        if not force_number or len(add_ind) == 0:
            return add_ind
        else:
            ti = K_best // len(add_ind)
            yu = K_best % len(add_ind)
            add_new = []
            for i in range(ti):
                add_new.extend(add_ind)
            add_new.extend(add_ind[:yu])
            return add_new


def Statis_func(stats=None):
    if stats is None:
        stats = {"fitness_dim_max": ("max",), "dim_is_target": ("sum",)}

    func = {"max": np.max, "mean": np.mean, "min": np.mean, "std": np.std, "sum": np.sum}
    att = {

        "fitness": lambda ind: ind.fitness.values[0],
        "fitness_dim_max": lambda ind: ind.fitness.values[0] if ind.dim_score else -np.inf,
        "fitness_dim_min": lambda ind: ind.fitness.values[0] if ind.dim_score else np.inf,
        "dim_is_target": lambda ind: 1 if ind.dim_score else 0,
        # special
        "coef": lambda ind: score_dim(ind.y_dim, "coef", fuzzy=False),
        "integer": lambda ind: score_dim(ind.y_dim, "integer", fuzzy=False),

        "length": lambda ind: len(ind),
        "height": lambda ind: ind.height,
        "h_GVP": lambda ind: (ind.height-1)/2,

        # mutil-target
        "weight_fitness": lambda ind: ind.fitness.wvalues,
        "weight_fitness_dim": lambda ind: ind.fitness.wvalues if ind.dim_score else -np.inf,
        # weight have mul the "-"
    }

    sa_all = {}

    for a, f in stats.items():
        if a in att:
            a_s = att[a]
        elif isinstance(callable, a):
            a_s = a
            a = str(a).split(" ")[1]
        else:
            raise TypeError("the key must be in definition or a function")
        sa = Statistics(a_s)
        if isinstance(f, str):
            f = [f, ]
        for i, fi in enumerate(f):
            assert fi in func
            ff = func[fi]

            sa.register(fi, ff)

        sa_all["Cal_%s" % a] = sa
    stats = MultiStatistics(sa_all)

    return stats


######################################
# shown                      #
######################################

def graph(expr):
    """Construct the graph of a tree expression. The tree expression must be
    valid. It returns in order a node list, an edge list, and a dictionary of
    the per node labels. The node are represented by numbers, the edges are
    tuples connecting two nodes (number), and the labels are values of a
    dictionary for which keys are the node numbers.

    :param expr: A tree expression to convert into a graph.
    :returns: A node list, an edge list, and a dictionary of labels.

    The returned objects can be used directly to populate a
    `pygraphviz <http://networkx.lanl.gov/pygraphviz/>`_ graph::

        import pygraphviz as pgv

        # [...] Execution of code that produce a tree expression

        nodes, edges, labels = graph(expr)

        g = pgv.AGraph()
        g.add_nodes_from(nodes)
        g.add_edges_from(edges)
        g.layout(prog="dot")

        for i in nodes:
            n = g.get_node(i)
            n.attr["label"] = labels[i]

        g.draw("tree.pdf")

    or a `NetworX <http://networkx.github.com/>`_ graph::

        import matplotlib.pyplot as plt
        import networkx as nx

        # [...] Execution of code that produce a tree expression

        nodes, edges, labels = graph(expr)

        g = nx.Graph()
        g.add_nodes_from(nodes)
        g.add_edges_from(edges)
        pos = nx.graphviz_layout(g, prog="dot")

        nx.draw_networkx_nodes(g, pos)
        nx.draw_networkx_edges(g, pos)
        nx.draw_networkx_labels(g, pos, labels)
        plt.show()


    .. note::

       We encourage you to use `pygraphviz
       <http://networkx.lanl.gov/pygraphviz/>`_ as the nodes might be plotted
       out of order when using `NetworX <http://networkx.github.com/>`_.
    """
    nodes = list(range(len(expr)))
    edges = list()
    labels = dict()

    stack = []
    for i, node in enumerate(expr):
        if stack:
            edges.append((stack[-1][0], i))
            stack[-1][1] -= 1
        labels[i] = repr(node)
        stack.append([i, node.arity])
        while stack and stack[-1][1] == 0:
            stack.pop()

    return nodes, edges, labels


def varAnd(population, toolbox, cxpb, mutpb):
    offspring = copy.deepcopy(population)

    # Apply crossover and mutation on the offspring
    for i in range(1, len(offspring), 2):
        if random.random() < cxpb:
            offspring[i - 1], offspring[i] = toolbox.mate(offspring[i - 1],
                                                          offspring[i])
            del offspring[i - 1].fitness.values, offspring[i].fitness.values

    for i in range(len(offspring)):
        if random.random() < mutpb:
            offspring[i], = toolbox.mutate(offspring[i])
            del offspring[i].fitness.values

    return offspring


def varAndfus(population, toolbox, cxpb, mutpb, fus):
    offspring = copy.deepcopy(population)

    # Apply crossover and mutation on the offspring
    for i in range(1, len(offspring), 2):
        if random.random() < cxpb:
            offspring[i - 1], offspring[i] = toolbox.mate(offspring[i - 1],
                                                          offspring[i])
            del offspring[i - 1].fitness.values, offspring[i].fitness.values

    mutpb /= len(fus)

    for j in fus:
        for i in range(len(offspring)):
            if random.random() < mutpb:
                # print(random.random(), i)
                offspring[i], = j(offspring[i])
                del offspring[i].fitness.values

    return offspring
